/***********************************************************


((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "layout.h"
#include "layout_int.h"

static int DefaultClose(LayoutObject plh);
static int dummy();
static int DefaultSetValues (LayoutObject   context,
                             LayoutValues   layout,
                             int          * index_returned);

static int DefaultGetValues (LayoutObject context,
                             LayoutValues layout,
                             int *index_returned);

static int wcsDefaultTransform (LayoutObject     context,
                                const wchar_t  * InpBuf,
                                size_t           InpSize,
                                void           * OutBuf,
                                size_t         * OutSize,
                                size_t         * ToOutBuf,
                                size_t         * ToInpBuf,
                                unsigned char  * BidiLvl,
                                size_t         * InpBufIndex);

static int DefaultTransform (LayoutObject    context,
                             const char    * InpBuf,
                             size_t          InpSize,
                             void          * OutBuf,
                             size_t        * OutSize,
                             size_t        * ToOutBuf,
                             size_t        * ToInpBuf,
                             unsigned char * BidiLvl,
                             size_t        * InpBufIndex);

static int inverse (char * buffer, size_t num_of_elements, int element_size);
/*----------------------------------------------------------------*/
LayoutObject DefaultOpen()
{
    LayoutObject plh = NULL;

    plh          = (LayoutObject) calloc(1, sizeof(LayoutObjectRec));
    if(!plh)
    {
        errno = ENOMEM;
        return((LayoutObject)NULL);
    }
    plh->methods = (LayoutMethods) calloc(1, sizeof(LayoutMethodsRec));
    if(!plh->methods)
    {
        errno = ENOMEM;
        DefaultClose(plh);
        return((LayoutObject)NULL);
    }
    /* set function addresses */
    plh->methods->create        = DefaultOpen;
    plh->methods->transform     = DefaultTransform;
    plh->methods->wcstransform  = wcsDefaultTransform;
    plh->methods->setvalues     = DefaultSetValues;
    plh->methods->getvalues     = DefaultGetValues;
    plh->methods->destroy       = DefaultClose;


    /* Bidirection and ShapeEditing not supported.  */
    plh->core.orientation = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(! plh->core.orientation)
    {
        errno = ENOMEM;
        DefaultClose(plh);
    }
    ((LayoutTextDescriptor) plh->core.orientation)->inp = ORIENTATION_LTR;
    ((LayoutTextDescriptor) plh->core.orientation)->out = ORIENTATION_LTR;
    plh->core.active_dir           = FALSE;      
    plh->core.active_shape_editing = FALSE;   

    return(plh);
}
/*----------------------------------------------------------------*/
static int DefaultClose(LayoutObject plh)
{
    if(plh)
    {
        if(plh->core.orientation)
            free(plh->core.orientation);
        if(plh->methods)
            free(plh->methods);
        free(plh);
        return(0);
    }
    else
        return(-1);   
}

/*----------------------------------------------------------------*/
/* A dummy function that returns 0. */
static int dummy()
{
   return(0);
}
/*----------------------------------------------------------------*/
static int DefaultSetValues (LayoutObject context,
                             LayoutValues layout,
                             int *index_returned)
{
    int                  i = 0;
    int                  Invalue;
    int                  Outvalue;
    LayoutTextDescriptor Descr;
 
    while (layout[i].name!=0)  /* while not end of array */
    {
        Descr=(LayoutTextDescriptor)layout[i].value;
        switch(layout[i].name)
        {
            case Orientation :
                /* extract needed value */
                Invalue  = Descr->inp;
                Outvalue = Descr->out;
                /* check if valid */
                if ((Invalue!=ORIENTATION_LTR)&&(Invalue!=ORIENTATION_RTL))
                    return(layout[i].name);
                if ((Outvalue!=ORIENTATION_LTR)&&(Outvalue!=ORIENTATION_RTL))
                    return(layout[i].name);
                /* assign value */
                ((LayoutTextDescriptor) context->core.orientation)->inp = Invalue;
                ((LayoutTextDescriptor) context->core.orientation)->out = Outvalue;
                break;
            case  Context           :
            case  TypeOfText        :
            case  ImplicitAlg       :
            case  Swapping          :
            case  Numerals          :
            case  TextShaping       :
            case  ShapeCharset      :
            case  InOutTextDescrMask:
            case  InOnlyTextDescr   :
            case  OutOnlyTextDescr  :
            case  CheckMode         :
                break;
            default:
            {
                *index_returned = i; 
                return(layout[i].name);
            }
        }
        i++;
    }
    return(0);
}

/*----------------------------------------------------------------*/
static int DefaultGetValues (LayoutObject context,
                             LayoutValues layout,
                             int *index_returned)
{
    int i=0;

    while (layout[i].name!=0)  /* while not end of array */
    {
        switch (layout[i].name)
        {
            case ActiveDirectional:
                *((BooleanValue *) (layout[i].value)) = context->core.active_dir;
                break;

            case ActiveShapeEditing:
                *((BooleanValue *) (layout[i].value)) = context->core.active_shape_editing;
                break;

            case Orientation:
                ((LayoutTextDescriptor) layout[i].value)->inp = ((LayoutTextDescriptor) context->core.orientation)->inp;
                ((LayoutTextDescriptor) layout[i].value)->out = ((LayoutTextDescriptor) context->core.orientation)->out;
                break;
            default: 
                *index_returned=i;
                return(layout[i].name);
                break;
        }
        i++;
    }
    return(0);
}
/*----------------------------------------------------------------*/
static int wcsDefaultTransform (LayoutObject context, 
                   const wchar_t  * InpBuf,
                   size_t           InpSize,
                   void           * OutBuf,
                   size_t         * OutSize,
                   size_t         * ToOutBuf,
                   size_t         * ToInpBuf,
                   unsigned char  * BidiLvl,
                   size_t         * InpBufIndex)
 {
    int       i;
    int       RC          = 0;
    size_t    temp_count;
    size_t    temp_index;
    size_t    to_copy;
    wchar_t * temp_buf;

    if(!InpBuf || !(*InpBuf))
    {
        errno = EINVAL;
        return(-1);
    }
    if(!OutBuf)
    {
        errno = EINVAL;
        return(-1);
    }
    if (OutSize == NULL || InpSize == 0 || *OutSize < 0)
    {
        errno = EINVAL;
        RC = -1;
        return (RC);
    }
    /* Decide on length of input string if not given */
    if (InpSize == (size_t) - 1 || InpSize > 10000)
    {
        i = 0;
        while (InpBuf[i] > (size_t) NULL)
        {
            if (i >= *OutSize && *OutSize)
            {
                RC = RC | 1;
                break;
            }
            i++;
        }
        InpSize = i;
    }
    if (*OutSize == 0)
    { /* calculate size of output buffer and return */
        *OutSize = InpSize;
    }
    if (InpSize > *OutSize)
    {
        InpSize = *OutSize;
        RC = RC | 1;
        errno = E2BIG;
    }
    /* copy input string to temporary buffer */
    temp_buf = (wchar_t *) calloc(InpSize, sizeof(wchar_t));
    if(!temp_buf)
    {
        errno = ENOMEM;
        return(-1);
    }
    memcpy( temp_buf, InpBuf, InpSize * (sizeof(wchar_t)) );
    temp_count = InpSize;
    temp_index = 0;

    if (ToOutBuf)             /* set maps to default values */
        for (i = 0; i < InpSize; i++)
            ToOutBuf[i] = i;
    if (ToInpBuf)
        for (i = 0; i < InpSize; i++)
            ToInpBuf[i] = i;
    if (BidiLvl)
        memset(BidiLvl, 0x80, InpSize);

                                     /* check if inversion is required */
    if (((LayoutTextDescriptor) context->core.orientation)->inp !=
        ((LayoutTextDescriptor) context->core.orientation)->out)
    {
        inverse((char *)temp_buf, InpSize, sizeof(wchar_t));  
        if (ToOutBuf)
            for (i = 0; i < InpSize; i++)
                ToOutBuf[i] = InpSize - 1 - ToOutBuf[i];
    } 

    /* copy temporary buffer to output buffer */
    if (OutBuf)
    {
        if (*OutSize > temp_count - temp_index)
            to_copy = temp_count - temp_index;
        else 
            to_copy = *OutSize;
        *OutSize = to_copy;
        memcpy(OutBuf, &(temp_buf[temp_index]), to_copy * sizeof(wchar_t)); /* just copy, no translation */
        temp_index = temp_index+to_copy;
        /* check if there was space for all the data */
        if (temp_index < temp_count)
            RC = E2BIG;
        else
        {
            free(temp_buf);
            temp_buf=NULL;
        }
    }
    return RC;      /* successfully performed */
}                  
/*| *********************************************************** */
/*| This is the old function of the m_wtransform.c              */
/*| *********************************************************** */
#ifdef INCLUDE_OLD_WTRANSFORM
{
    char * mbsin;
    char * mbsout;
    int    mlenin;
    int    mlenout;
    int    RC;
    int    i;


    if(!InpBuf || !(*InpBuf))
    {
        errno = EINVAL;
        return(-1);
    }
    if(!OutBuf)
    {
        errno = EINVAL;
        return(-1);
    }
    if (OutSize == NULL || InpSize == 0 || *OutSize < 0)
    {
        errno = EINVAL;
        RC = -1;
        return (RC);
    }
    /* Decide on length of input string if not given */
    if (InpSize == (size_t) - 1 || InpSize > 10000)
    {
        i = 0;
        while (InpBuf[i] > (size_t) NULL)
        {
            if (i >= *OutSize && *OutSize)
            {
                RC = RC | 1;
                break;
            }
            i++;
        }
        InpSize = i;
    }
    if (*OutSize == 0)
    { /* calculate size of output buffer and return */
        *OutSize = InpSize;
    }
    if (InpSize > *OutSize)
    {
        InpSize = *OutSize;
        RC = RC | 1;
        errno = E2BIG;
    }
    
    mbsin  = (char *) calloc(InpSize + 1,  MB_CUR_MAX);  /* enough memory */
    if(!mbsin)
    {
        errno = ENOMEM;
        return(-1);
    }
    mlenin = InpSize * MB_CUR_MAX;
    wcstombs(mbsin, InpBuf, mlenin);

    mbsout = calloc(*OutSize + 1, MB_CUR_MAX); 
    if(!mbsout)
    {
        errno = ENOMEM;
        free(mbsin);
        return(-1);
    }
    mlenout = *OutSize * MB_CUR_MAX;

    /* send multi-byte string */
    RC = DefaultTransform(plh,
                          mbsin,     mlenin,
                          mbsout,    &mlenout,
                          &ToOutBuf, &ToInpBuf,
                          &BidiLvl,  &plh);

    *OutSize = mbstowcs(OutBuf, mbsout, mlenout);
    return(RC);
}
#endif /* INCLUDE_OLD_WTRANSFORM */
/*----------------------------------------------------------------*/



static int inverse (char * buffer, size_t num_of_elements, int element_size)
/* inverse a stream of elements, where each elemnt is of size element_size */
{
    char  * temp; 
    int     i, j;
    int     start, end;

    temp = (char *) calloc (element_size + 1, sizeof(char));
    if(!temp)
    {
        errno = ENOMEM;
        return(-1);
    }
    for (i = 0, j = num_of_elements - 1; i < j; i++, j--)
    {
        memcpy(temp, &(buffer[i * element_size]), element_size);
        memcpy(&(buffer[i * element_size]), &(buffer[j * element_size]), element_size);
        memcpy(&(buffer[j * element_size]), temp, element_size);
    }
    free (temp);
    return(0);
}
/*----------------------------------------------------------------*/
static int DefaultTransform (LayoutObject    context,
                             const char    * InpBuf,
                             size_t          InpSize,
                             void          * OutBuf,
                             size_t        * OutSize,
                             size_t        * ToOutBuf,
                             size_t        * ToInpBuf,
                             unsigned char * BidiLvl,
			     size_t        * InpBufIndex)
                   
{
    int       i;
    int       RC          = 0;
    size_t    temp_count;
    size_t    temp_index;
    size_t    to_copy;
    char    * temp_buf;

    if(!InpBuf || !(*InpBuf))
    {
        errno = EINVAL;
        return(-1);
    }
    if(!OutBuf)
    {
        errno = EINVAL;
        return(-1);
    }
    if (OutSize == NULL || InpSize == 0 || *OutSize < 0)
    {
        errno = EINVAL;
        RC = -1;
        return (RC);
    }
    /* Decide on length of input string if not given */
    if (InpSize == (size_t) - 1 || InpSize > 10000)
    {
        i = 0;
        while (InpBuf[i] > (size_t) NULL)
        {
            if (i >= *OutSize && *OutSize)
            {
                RC = RC | 1;
                break;
            }
            i++;
        }
        InpSize = i;
    }
    if (*OutSize == 0)
    { /* calculate size of output buffer and return */
        *OutSize = InpSize;
    }
    if (InpSize > *OutSize)
    {
        InpSize = *OutSize;
        RC = RC | 1;
        errno = E2BIG;
    }
    /* copy input string to temporary buffer */
    temp_buf = (char *) calloc(InpSize, sizeof(char));
    if(!temp_buf)
    {
        errno = ENOMEM;
        return(-1);
    }
    memcpy(temp_buf, InpBuf, InpSize);
    temp_count=InpSize;
    temp_index=0;

    if (ToOutBuf)             /* set maps to default values */
        for (i = 0; i < InpSize; i++)
            ToOutBuf[i] = i;
    if (ToInpBuf)
        for (i = 0; i < InpSize; i++)
            ToInpBuf[i] = i;
    if (BidiLvl)
        memset(BidiLvl, 0x80, InpSize);

                                     /* check if inversion is required */
    if (((LayoutTextDescriptor) context->core.orientation)->inp !=
        ((LayoutTextDescriptor) context->core.orientation)->out)
    {
        inverse(temp_buf, InpSize, sizeof(char));  
     /*   if (ToInpBuf)
            inverse(ToInpBuf, InpSize, sizeof(unsigned int));*/
        if (ToOutBuf)
            for (i = 0; i < InpSize; i++)
                ToOutBuf[i] = InpSize - 1 - ToOutBuf[i];
    } 

    /* copy temporary buffer to output buffer */
    if (OutBuf)
    {
        if (*OutSize > temp_count - temp_index)
            to_copy = temp_count - temp_index;
        else 
            to_copy = *OutSize;
        *OutSize = to_copy;
        memcpy(OutBuf, &(temp_buf[temp_index]), to_copy); /* just copy, no translation */
        temp_index = temp_index+to_copy;
        /* check if there was space for all the data */
        if (temp_index < temp_count)
            RC = E2BIG;
        else
        {
            free(temp_buf);
            temp_buf=NULL;
        }
    }
    return RC;      /* successfully performed */
}
/*----------------------------------------------------------------*/
