/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group

******************************************************************/


#ifndef _Lc_LayoutP_
#define _Lc_LayoutP_

#include <sys/lc_core.h>

#define _LC_LAYOUT 1

/******************************************************************************/
/*  Structure LayoutObjectP contains variables to hold the addresses of       */
/*  all the API calls that are exported by libi18n.a .                        */
/*  The initialization routine of the library (LayoutOpen) loads              */
/*  the relevant  module from LOCPATH, depending on the                       */
/*  locale specified by the application programmer, then it executes          */
/*  the initialization routine of that module. (The address of	              */
/*  the initilization routine of the module is found in variable              */
/*  "init" of structure LC_struct that is returned from lc_load.)             */
/*  The initilization routine of the module allocates a structure             */
/*  of type La, fills it and returns a pointer to it. This pointer            */
/*  must be passed to all subsequent API calls. Finally a call to             */
/*  LayoutClose will free this structure.                                     */
/*  The initialization routine of the module fills the LayoutObjectP          */
/*  structure as follows : It sets the address of all the API calls, so that  */
/*  later when an API call is requested from the library, all the library does*/
/*  is to call the address of that function.                                  */
/******************************************************************************/
typedef struct __lc_layout_methods {
    struct __lc_layout *(*Open)();            
    int (*Transform)();
    int (*EditShape)();
    int (*wcsTransform)();
    int (*wcsEditShape)();
    int (*ShapeBoxChars)();
    int (*SetValues)();
    int (*GetValues)();
    int (*Close)();            
    } LayoutMethodsRec, *LayoutMethods;


typedef struct {
    struct _lc_layout* lo_class;
    BooleanValue Bidirection;
    BooleanValue ShapeEditing;
    char *Values;
    _LC_object_t *locale;
    char *extension;
    } LayoutObjectCoreRec, *LayoutObjectCore;

typedef struct {
    LayoutObjectCore core;
    LayoutMethods methods;
    }LayoutObjectPRec, *LayoutObjectP;

typedef struct _lc_layout{
    _LC_object_t lc_hdr; 
    LayoutObjectP (*initialize)();            
} lc_layout;

#ifdef _NO_PROTO
extern int layout_object_create();
extern int layout_object_free();
extern int layout_object_transform();
extern int wcslayout_object_transform();
extern int layout_object_editshape();
extern int wcslayout_object_editshape ();
extern int layout_object_shapeboxchars();
extern int layout_object_setvalue();
extern int layout_object_getvalue();
#else
extern int layout_object_create(const char*, LayoutObjectP*);
extern int layout_object_free(LayoutObjectP);
extern int layout_object_transform (LayoutObjectP, const char*, size_t*, void*, size_t*, size_t*, size_t*, unsigned char *);
extern int wcslayout_object_transform (LayoutObjectP, const wchar_t*, size_t*, void*, size_t*, size_t*, size_t*, unsigned char *);
extern int layout_object_editshape (LayoutObjectP, BooleanValue, size_t*, const char *, size_t*, void*, size_t*);
extern int wcslayout_object_editshape (LayoutObjectP, BooleanValue, size_t*, const wchar_t*, size_t*, void*, size_t*);
extern int layout_object_shapeboxchars(LayoutObjectP, const char*, const size_t, char *);
extern int layout_object_setvalue (LayoutObjectP, LayoutValues, int *);
extern int layout_object_getvalue (LayoutObjectP, LayoutValues, int *);
#endif /* _NO_PROTO */


#endif /*_Lc_LayoutP_ */
