/***********************************************************


((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

				
#include <stdio.h>
#include <stdlib.h>
#include <langinfo.h>
#include <iconv.h>
#include <errno.h>
#include "../layout/layout.h"
#include "layout_int.h"
#include "uc_bdstruct.h"
#include "VNlayout.h"           /* VNnoSpace[],  VNcomp[][3]... */

unsigned short 
VNcomposedTone(unsigned short VNbaseChar,
               unsigned short VNToneMark)
{
    int             i;

    if (VNToneMark == 0x0340)
        VNToneMark = 0x0300;
    if (VNToneMark == 0x0341)
        VNToneMark = 0x0301;

    for (i = 0; i < VNNumOfCompToneSet; i++) {
        if (VNbaseChar == VNcompTone[i][0] && VNToneMark == VNcompTone[i][1])
            return (VNcompTone[i][2]);
    }

    return (0);                 /* Error compose Seq */
};


unsigned short 
VNcomposedDiac(unsigned short VNbaseChar,
               unsigned short VNDiacMark)
{
    int             i;

    for (i = 0; i < VNNumOfCompDiacSet; i++) {
        if (VNbaseChar == VNcompDiac[i][2] && VNDiacMark == VNcompDiac[i][1])
            return (VNcompDiac[i][2]);
    }

    return (0);                 /* Error compose Seq */
};

int 
VNTransform(
            PICSPARAMRec IcsRec,/* Not Used in Vietnamese */
            const unsigned short * InpBuf,      /* UCS2 input chars */
            const size_t InpSize,       /* # of UCS2 input chars */
            unsigned short * OutBuf,    /* UCS2 output chars */
            size_t * OutSize,   /* # of UCS2 output chars */
            size_t * InpToOut,  /* in to out map (=InSize) */
            size_t * OutToInp,  /* out to in map (=OutSize) */
            unsigned char *Property)
{                               /* Compose mark (=InSize) */
    unsigned short         ret, Last;
    int             i, j, k;
    int             Found;
    unsigned short        *inpBuf;

    if (InpSize == 0) {         /* Error! Zero Input Buffer */
        errno = EINVAL;
        return (-1);
    }
    inpBuf = (unsigned short *) calloc(InpSize, sizeof(unsigned short));
    if(!inpBuf)
    {
        errno = ENOMEM;
        return(-1);
    }
    for (i = 0; i < InpSize; i++)
        inpBuf[i] = InpBuf[i];

    *OutSize = 0;
    Last = LastNormal;

    for (j = 0; j < VNNumOfNoSpace && Last != LastToneMark; j++) {
        if (*inpBuf == VNnoSpaceTone[j]) {
            /* Error??? Tone mark at 1st position of the buffer. */
            /* retrun a has space mark.                          */
            *OutBuf = VNhasSpaceTone[j];
            *InpToOut = 0;
            *OutToInp = 0;
            *Property += 0x80;
            *OutSize += 1;
            Last = LastToneMark;
        }
    }

    if (Last != LastToneMark) {
        /* return as normal */
        *OutBuf = *inpBuf;
        *InpToOut = 0;
        *OutToInp = 0;
        *Property += 0x80;
        *OutSize += 1;
        Last = LastNormal;
    }
    for (i = 1; i < InpSize; i++) {
        Found = NotFoundCompose;

        /* Find diacritical mark */
        for (j = 0; j < VNNumOfDiacritical && Found == NotFoundCompose; j++) {
            if (*(inpBuf + i) == VNdiacritical[j]) {
                Found = FoundCompose;
                ret = VNcomposedDiac(*(inpBuf + i - 1), *(inpBuf + i));
                if (ret != 0) { /* found correct diac combining */
                    *(OutBuf + (*OutSize - 1)) = ret;
                    *(InpToOut + i) = *OutSize - 1;
                    /* *OutToInp no changes */
                    *(Property + i) += 0;
                    /* OutSize no changes */
                    Found = CorrectCompose;
                    Last = LastDiacMark;
                }
            }
        }

        /* Find no spacing tone mark char */
        for (j = 0; j < VNNumOfNoSpace && Found == NotFoundCompose; j++) {
            if (*(inpBuf + i) == VNnoSpaceTone[j]) {
                Found = FoundCompose;
                ret = VNcomposedTone(*(OutBuf + (*OutSize - 1)), *(inpBuf + i));
                if (ret != 0) { /* found correct Tone combining */
                    *(OutBuf + (*OutSize - 1)) = ret;
                    *(InpToOut + i) = *OutSize - 1;
                    /* *OutToInp no changes */
                    *(Property + i) += 0;
                    /* OutSize no changes */
                    Found = CorrectCompose;
                    Last = LastToneMark;
                }
            }
        }

        /* Not Found correct Tone Mark or Diacritical Mark Compose */
        if (Found != CorrectCompose) {
            *(OutBuf + *OutSize) = *(inpBuf + i);
            *(InpToOut + i) = *OutSize;
            *(OutToInp + (unsigned short) *OutSize) = i;
            *(Property + i) += 0x80;
            *OutSize += 1;
            Last = LastNormal;
        }
    }
    free(inpBuf);
    return (0);
};
