/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group

******************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>

#define VNNumOfNoSpace          7
#define VNNumOfCompToneSet      121
#define VNNumOfVowle            24
#define VNNumOfDiacritical      7
#define VNNumOfCompDiacSet      30

#define LastNormal              1
#define LastToneMark            2
#define LastDiacMark            3

#define NotFoundCompose         0
#define FoundCompose            1
#define CorrectCompose          2

static unsigned short  VNnoSpaceTone[VNNumOfNoSpace] = {
                                                 0x0300,        /* ' */
                                                 0x0301,        /* ` */
                                                 0x0303,        /* ~ */
                                                 0x0309,        /* ? */
                                                 0x0323,        /* . */
                                                 0x0340,        /* ` */
                                                 0x0341         /* ' */
};

static unsigned short  VNhasSpaceTone[VNNumOfNoSpace] = {
                                                  0x0060,       /* ' */
                                                  0x0027,       /* ` */
                                                  0x007E,       /* ~ */
                                                  0x003F,       /* ? */
                                                  0x002E,       /* . */
                                                  0x0027,       /* ` */
                                                  0x0060        /* ' */
};

static unsigned short  VNvowle[VNNumOfVowle] = {
                                         0x0041,        /* A  */
                                         0x0061,        /* a  */
                                         0x0102,        /* A( */
                                         0x0103,        /* a( */
                                         0x00C2,        /* A^ */
                                         0x00E2,        /* a^ */
                                         0x0045,        /* E  */
                                         0x0065,        /* e  */
                                         0x00CA,        /* E^ */
                                         0x00EA,        /* e^ */
                                         0x0049,        /* I  */
                                         0x0069,        /* i  */
                                         0x004F,        /* O  */
                                         0x006F,        /* o  */
                                         0x00D4,        /* O^ */
                                         0x00F4,        /* o^ */
                                         0x01A0,        /* O" */
                                         0x01A1,        /* o" */
                                         0x0055,        /* U  */
                                         0x0075,        /* u  */
                                         0x01AF,        /* U" */
                                         0x01B0,        /* u" */
                                         0x0059,        /* Y  */
                                         0x0079         /* y  */
};

static unsigned short  VNdiacritical[] = {
                                   0x0306,      /* ( Breve      */
                                   0x0302,      /* ^ Circumflex */
                                   0x0315,      /* " Horn       */
                                   0x0308,      /* : Diaeresis  */
                                   0x030A,      /* 0 Ring-Above */
                                   0x0338,      /* / Slash      */
                                   0x0339       /* , Cedilla    */
};

static unsigned short  VNcompDiac[VNNumOfCompDiacSet][3] = {
                                                 /* Base,   Comp,   Combined    comment */
                                                   {0x0041, 0x0306, 0x0102},    /* A( */
                                                   {0x0061, 0x0306, 0x0103},    /* a( */
                                                   {0x0041, 0x0302, 0x00C2},    /* A^ */
                                                   {0x0061, 0x0302, 0x00E2},    /* a^ */
                                                   {0x0041, 0x0308, 0x00C4},    /* A: */
                                                   {0x0061, 0x0308, 0x00E4},    /* a: */
                                                   {0x0041, 0x030A, 0x00C5},    /* Ao */
                                                   {0x0061, 0x030A, 0x00E5},    /* ao */
                                                   {0x0045, 0x0302, 0x00CA},    /* E^ */
                                                   {0x0061, 0x0302, 0x00EA},    /* e^ */
                                                   {0x0045, 0x0308, 0x00CB},    /* E: */
                                                   {0x0061, 0x0308, 0x00EB},    /* e: */
                                                   {0x0049, 0x0302, 0x00CE},    /* I^ */
                                                   {0x0069, 0x0302, 0x00EE},    /* i^ */
                                                   {0x0049, 0x0308, 0x00CF},    /* I: */
                                                   {0x0069, 0x0308, 0x00EF},    /* i: */
                                                   {0x004F, 0x0302, 0x00D4},    /* O^ */
                                                   {0x006F, 0x0302, 0x00F4},    /* o^ */
                                                   {0x004F, 0x0308, 0x00D6},    /* O: */
                                                   {0x006F, 0x0308, 0x00F6},    /* o: */
                                                   {0x004F, 0x0315, 0x01A0},    /* O" */
                                                   {0x006F, 0x0315, 0x01A1},    /* o" */
                                                   {0x004F, 0x0338, 0x00D8},    /* O/ */
                                                   {0x006F, 0x0338, 0x00F8},    /* o/ */
                                                   {0x0055, 0x0302, 0x00DB},    /* U^ */
                                                   {0x0075, 0x0302, 0x00FB},    /* u^ */
                                                   {0x0055, 0x0315, 0x00DC},    /* U" */
                                                   {0x0075, 0x0315, 0x00FC},    /* u" */
                                                   {0x0059, 0x0308, 0x0178},    /* Y: */
                                                   {0x0079, 0x0308, 0x00FF},    /* y: */
};

static unsigned short  VNcompTone[VNNumOfCompToneSet][3] = {
                                                 /* Base,   Comp,   Combined    comment */
                                                   {0x0061, 0x0300, 0x00E0},    /* a`  */
                                                   {0x0041, 0x0300, 0x00C0},    /* A`  */
                                                   {0x0061, 0x0309, 0x1EA3},    /* a?  */
                                                   {0x0041, 0x0309, 0x1EA2},    /* A?  */
                                                   {0x0061, 0x0303, 0x00E3},    /* a~  */
                                                   {0x0041, 0x0303, 0x00C3},    /* A~  */
                                                   {0x0061, 0x0301, 0x00E1},    /* a'  */
                                                   {0x0041, 0x0301, 0x00C1},    /* A'  */
                                                   {0x0061, 0x0323, 0x1EA1},    /* a.  */
                                                   {0x0041, 0x0323, 0x1EA0},    /* A.  */
                                                   {0x0103, 0x0300, 0x1EB1},    /* a(` */
                                                   {0x0102, 0x0300, 0x1EB0},    /* A(` */
                                                   {0x0103, 0x0309, 0x1EB3},    /* a(? */
                                                   {0x0102, 0x0309, 0x1EB2},    /* A(? */
                                                   {0x0103, 0x0303, 0x1EB5},    /* a(~ */
                                                   {0x0102, 0x0303, 0x1EB4},    /* A(~ */
                                                   {0x0103, 0x0301, 0x1EAF},    /* a(' */
                                                   {0x0102, 0x0301, 0x1EAE},    /* A(' */
                                                   {0x0103, 0x0323, 0x1EB7},    /* a(. */
                                                   {0x0102, 0x0323, 0x1EB6},    /* A(. */
                                                   {0x00E2, 0x0300, 0x1EA7},    /* a^` */
                                                   {0x00C2, 0x0300, 0x1EA6},    /* A^` */
                                                   {0x00E2, 0x0309, 0x1EA9},    /* a^? */
                                                   {0x00C2, 0x0309, 0x1EA8},    /* A^? */
                                                   {0x00E2, 0x0303, 0x1EAB},    /* a^~ */
                                                   {0x00C2, 0x0303, 0x1EAA},    /* A^~ */
                                                   {0x00E2, 0x0301, 0x1EA5},    /* a^' */
                                                   {0x00C2, 0x0301, 0x1EA4},    /* A^' */
                                                   {0x00E2, 0x0323, 0x1EAD},    /* a^. */
                                                   {0x00C2, 0x0323, 0x1EAC},    /* A^. */
                                                   {0x0065, 0x0300, 0x00E8},    /* e`  */
                                                   {0x0045, 0x0300, 0x00C8},    /* E`  */
                                                   {0x0065, 0x0309, 0x1EBB},    /* e?  */
                                                   {0x0045, 0x0309, 0x1EBA},    /* E?  */
                                                   {0x0065, 0x0303, 0x1EBD},    /* e~  */
                                                   {0x0045, 0x0303, 0x1EBC},    /* E~  */
                                                   {0x0065, 0x0301, 0x00E9},    /* e'  */
                                                   {0x0045, 0x0301, 0x00C9},    /* E'  */
                                                   {0x0065, 0x0323, 0x1EB9},    /* e.  */
                                                   {0x0045, 0x0323, 0x1EB8},    /* E.  */
                                                   {0x00EA, 0x0300, 0x1EC1},    /* e^` */
                                                   {0x00CA, 0x0300, 0x1EC0},    /* E^` */
                                                   {0x00EA, 0x0309, 0x1EC3},    /* e^? */
                                                   {0x00CA, 0x0309, 0x1EC2},    /* E^? */
                                                   {0x00EA, 0x0303, 0x1EC5},    /* e^~ */
                                                   {0x00CA, 0x0303, 0x1EC4},    /* E^~ */
                                                   {0x00EA, 0x0301, 0x1EBF},    /* e^~ */
                                                   {0x00CA, 0x0301, 0x1EBE},    /* E^~ */
                                                   {0x00EA, 0x0323, 0x1EC7},    /* e^. */
                                                   {0x00CA, 0x0323, 0x1EC6},    /* E^. */
                                                   {0x0069, 0x0300, 0x00EC},    /* i`  */
                                                   {0x0049, 0x0300, 0x00CC},    /* I`  */
                                                   {0x0069, 0x0309, 0x1EC9},    /* i?  */
                                                   {0x0049, 0x0309, 0x1EC8},    /* I?  */
                                                   {0x0069, 0x0303, 0x0129},    /* i~  */
                                                   {0x0049, 0x0303, 0x0129},    /* I~  */
                                                   {0x0069, 0x0301, 0x00ED},    /* i'  */
                                                   {0x0049, 0x0301, 0x00CD},    /* I'  */
                                                   {0x0069, 0x0323, 0x1ECB},    /* i.  */
                                                   {0x0049, 0x0323, 0x1ECA},    /* I.  */
                                                   {0x006F, 0x0300, 0x00F2},    /* o`  */
                                                   {0x004F, 0x0300, 0x00D2},    /* O`  */
                                                   {0x006F, 0x0309, 0x1ECF},    /* o?  */
                                                   {0x004F, 0x0309, 0x1ECE},    /* O?  */
                                                   {0x006F, 0x0303, 0x00F5},    /* o~  */
                                                   {0x004F, 0x0303, 0x00D5},    /* O~  */
                                                   {0x006F, 0x0301, 0x00F3},    /* o'  */
                                                   {0x004F, 0x0301, 0x00D3},    /* O'  */
                                                   {0x006F, 0x0323, 0x1ECD},    /* o.  */
                                                   {0x004F, 0x0323, 0x1ECC},    /* O.  */
                                                   {0x00F4, 0x0300, 0x1ED3},    /* o^` */
                                                   {0x00D4, 0x0300, 0x1ED3},    /* O^` */
                                                   {0x00F4, 0x0309, 0x1ED5},    /* o^? */
                                                   {0x00D4, 0x0309, 0x1ED4},    /* O^? */
                                                   {0x00F4, 0x0303, 0x1ED7},    /* o^~ */
                                                   {0x00D4, 0x0303, 0x1ED6},    /* O^~ */
                                                   {0x00F4, 0x0301, 0x1ED1},    /* o^' */
                                                   {0x00D4, 0x0301, 0x1ED0},    /* O^' */
                                                   {0x00F4, 0x0323, 0x1ED9},    /* o^. */
                                                   {0x00D4, 0x0323, 0x1ED8},    /* O^. */
                                                   {0x01A1, 0x0300, 0x1EDD},    /* o"` */
                                                   {0x01A0, 0x0300, 0x1EDC},    /* O"` */
                                                   {0x01A1, 0x0309, 0x1EDF},    /* o"? */
                                                   {0x01A0, 0x0309, 0x1EDE},    /* O"? */
                                                   {0x01A1, 0x0303, 0x1EE1},    /* o"~ */
                                                   {0x01A0, 0x0303, 0x1EE0},    /* O"~ */
                                                   {0x01A1, 0x0301, 0x1EDB},    /* o"' */
                                                   {0x01A0, 0x0301, 0x1EDA},    /* O"' */
                                                   {0x01A1, 0x0323, 0x1EE3},    /* o". */
                                                   {0x01A0, 0x0323, 0x1EE2},    /* O". */
                                                   {0x0075, 0x0300, 0x00F9},    /* u`  */
                                                   {0x0055, 0x0300, 0x00D9},    /* U`  */
                                                   {0x0075, 0x0309, 0x1EE7},    /* u?  */
                                                   {0x0055, 0x0309, 0x1EE6},    /* U?  */
                                                   {0x0075, 0x0303, 0x0169},    /* u~  */
                                                   {0x0055, 0x0303, 0x0168},    /* U~  */
                                                   {0x0075, 0x0301, 0x00FA},    /* u'  */
                                                   {0x0055, 0x0301, 0x00DA},    /* U'  */
                                                   {0x0075, 0x0323, 0x1EE5},    /* u.  */
                                                   {0x0055, 0x0323, 0x1EE4},    /* U.  */
                                                   {0x01B0, 0x0300, 0x1EEB},    /* u"` */
                                                   {0x01AF, 0x0300, 0x1EEA},    /* U"` */
                                                   {0x01B0, 0x0309, 0x1EED},    /* u"? */
                                                   {0x01AF, 0x0309, 0x1EEC},    /* U"? */
                                                   {0x01B0, 0x0303, 0x1EEF},    /* u"~ */
                                                   {0x01AF, 0x0303, 0x1EEE},    /* U"~ */
                                                   {0x01B0, 0x0301, 0x1EE9},    /* u"' */
                                                   {0x01AF, 0x0301, 0x1EE8},    /* U"' */
                                                   {0x01B0, 0x0323, 0x1EF1},    /* u". */
                                                   {0x01AF, 0x0323, 0x1EF0},    /* U". */
                                                   {0x0079, 0x0300, 0x1EF3},    /* y`  */
                                                   {0x0059, 0x0300, 0x1EF2},    /* Y`  */
                                                   {0x0079, 0x0309, 0x1EF7},    /* y?  */
                                                   {0x0059, 0x0309, 0x1EF6},    /* Y?  */
                                                   {0x0079, 0x0303, 0x1EF9},    /* y~  */
                                                   {0x0059, 0x0303, 0x1EF8},    /* Y~  */
                                                   {0x0079, 0x0301, 0x00FD},    /* y'  */
                                                   {0x0059, 0x0301, 0x00DD},    /* Y'  */
                                                   {0x0079, 0x0323, 0x1EF5},    /* y.  */
                                                   {0x0059, 0x0323, 0x1EF4},    /* Y.  */
                                                   {0x0000, 0x0000, 0x0000}     /* END */
};
