/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#ifndef _ARABIC_H_
#define _ARABIC_H_
#include <iconv.h>
#include <stdlib.h>
#include "../layout/layout.h"

#define ArabicSpecialShaping            (1L << 26)
#define ArabicOneCellShaping            (1L << 27)
#define WordBreak                       (1L << 28)

#define MaskArabic                      0xf0000000
#define MaskSpecialShaping              0x30000000
#define MaskOneCellShaping              0x40000000
#define MaskWordBreak                   0x80000000

/* for the ArabicSpecialShaping */
#define TEXT_STANDARD                   0x00000000
#define TEXT_SPECIAL                    0x10000000
#define TEXT_COMPOSED                   0x20000000

/* For the ArabicOneCellShaping */
#define ONECELL_SEEN                    0x40000000
#define TWOCELL_SEEN                    0x00000000

/* For the WordBreak */
#define BREAK                           0x80000000
#define NO_BREAK                        0x00000000

#define EDITINPUT                       0
#define EDITREPLACE                     1

/* 
 * Define Layout Values used for LayoutSet/GetValues()
 */
#define _AllTextDescptors               0x0000ffff
#define _Orientation                    0x00000001
#define _TypeOfText                     0x00000002
#define _Swapping                       0x00000004
#define _Numerals                       0x00000008
#define _TextShaping                    0x00000010
#define _ArabicSpecialShaping           0x00000020
#define _ArabicOneCellShaping           0x00000040
#define _WordBreak                      0x00000080
#define _BidiType                       0x00000100

#define _ActiveShapeEditing             (0x01<<16)
#define _ActiveBidirection              (0x02<<16)
#define _ShapeCharset                   (0x03<<16)
#define _ShapeCharsetSize               (0x04<<16)
#define _ShapeContextSize               (0x05<<16)
#define _InputMode                      (0x07<<16)

#define _TEXT_VISUAL                    0x00000000
#define _TEXT_IMPLICIT                  0x01000000
#define _TEXT_EXPLICIT                  0x02000000

#define _BIDI_DEFAULT                   0x00000000
#define _BIDI_UCS                       0x04000000

#define _ORIENTATION_LTR                0x00000000
#define _ORIENTATION_RTL                0x00010000
#define _ORIENTATION_CONTEXT_LTR        0x00020000
#define _ORIENTATION_CONTEXT_RTL        0x00030000

#define _NUMERALS_NOMINAL               0x00000000
#define _NUMERALS_NATIONAL              0x00002000
#define _NUMERALS_CONTEXTUAL            0x00003000

#define _BREAK                          0x00000200
#define _NO_BREAK                       0x00000000

#define _SWAPPING                       0x00000100
#define _NO_SWAPPING                    0x00000000

#define _ONECELL_SEEN                   0x00000080
#define _TWOCELL_SEEN                   0x00000000

#define _TEXT_STANDARD                  0x00000000
#define _TEXT_SPECIAL                   0x00000040
#define _TEXT_COMPOSED                  0x00000020

#define _TEXT_SHAPED                    0x00000000
#define _TEXT_NOMINAL                   0x00000010
#define _TEXT_INITIAL                   0x00000011
#define _TEXT_MIDDLE                    0x00000012
#define _TEXT_FINAL                     0x00000013
#define _TEXT_ISOLATED                  0x00000014
 
#define _EDITINPUT                      0
#define _EDITREPLACE                    1

/* This part is private to the Arabic part and is added for downword */
/* compatibility with previous versions.                             */
typedef struct {
    LayoutTextDescriptor OneCell;
    LayoutTextDescriptor Word;
    LayoutTextDescriptor SpecialSh;
    iconv_t              iconv_handle;
    char                 *tempbuf;
    size_t               temp_count;
    size_t               temp_index;
    char                 *InCharSet;
    char                 ShapeState;
    /* The following pointer is for the UNICODE data */
    void                 *uni_data;
}   ArabicRec, *Arabic;

#endif
