/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/
                                             
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <errno.h>
#include "mainstruct.h"
#include "arabic.h"
#include "uc_data.h"
#include "layout_backward.h"
#include "layout_int.h"

LayoutObject    BidiCreate();
static int      BidiClose();

extern int      BidiTransform();
extern int      wcsBidiTransform();
//extern int      BidiEditShape();
extern int      BidiSetValues();
extern int      BidiGetValues();

/***********************************************************************/
/* FUNCTION NAME: BidiClose()                                          */
/*                                                                     */
/* FUNCTION DESCRIPTION: This routine is used to free up the allocated */
/*                       memory for the LayoutObject structure and the */
/*                       private data.                                 */
/***********************************************************************/
static int 
BidiClose(plh)
LayoutObj plh;
{
    /* BidiValuesRec bidivalues; */
    Default         defaultData;
    Arabic          arabData;
    UbData          uniData;

    defaultData   = plh->private_data;
    arabData      = defaultData->locale_data;
    uniData       = arabData->uni_data;

    if (uniData->iconv_handle_in != (iconv_t)-1){
        if (!uniData->iconv_handle_in)
            ;
        else
            iconv_close(uniData->iconv_handle_in);
    }
    if (uniData->iconv_handle_out != (iconv_t)-1){
        if (!uniData->iconv_handle_out)
            ;
        else
            iconv_close(uniData->iconv_handle_out);
    }
    free(uniData);
    free(arabData);
    free(defaultData->methods);
    free(defaultData);

    free(plh->methods);
    if (plh->core.orientation)
        free(plh->core.orientation);

    if (plh->core.context)
        free(plh->core.context);

    if (plh->core.type_of_text)
        free(plh->core.type_of_text);

    if (plh->core.implicit_alg)
        free(plh->core.implicit_alg);

    if (plh->core.swapping)
        free(plh->core.swapping);

    if (plh->core.numerals)
        free(plh->core.numerals);

    if (plh->core.text_shaping)
        free(plh->core.text_shaping);

    if (plh->core.shape_context_size)
        free(plh->core.shape_context_size);

    if (plh->core.shape_charset)
        free(plh->core.shape_charset);

    if (plh->core.locale_name)
        free(plh->core.locale_name);

    free(plh);
    return (0);
}

/***********************************************************************/
/* FUNCTION NAME: SetCurrentCodeSet()                                  */
/*                                                                     */
/* FUNCTION DESCRIPTION: This function returns the current code set.   */
/***********************************************************************/
static int SetCurrentCodeSet(values)
LayoutCore values;
{
    char           *cp;

    cp = (char *) nl_langinfo(CODESET);
    values->shape_charset = (char *) calloc(strlen(cp) + 1, sizeof(char));

    if(!values->shape_charset) {
        errno = ENOMEM;
	return(-1);
    }
    if (strcmp(cp, ARABIC_CHARSET2) == 0) {
        cp = ARABIC_CHARSET1;
    }

    strcpy(values->shape_charset, cp);
    return(0);
}

/*********************************************************************/
/* FUNCTION NAME: _LayoutObjectInit()                                */
/*                                                                   */
/* FUNCTION DESCRIPTION: This is the BIDI library initialization     */
/*                       routine. It creates the LayoutObject struct.*/
/*                       and fills it with the addresses of the      */
/*                       functions to be called to handle the        */
/*                       processing of the corresponding call to any */
/*                       of the PLS X/Open routines "m_*_layout()"   */
/*                       and sets the default values in the          */
/*                       LayoutObject structure.                     */
/*********************************************************************/
LayoutObject _LayoutObjectInit(char *locale_name)
{
    LayoutObj       plh;
    int             rc = 0;
    int             j  = 0;
    LayoutValues    layoutValues = (LayoutValues) NULL;
    LayoutCore      bidivalues;
    Default         defaultData;
    Arabic          arabicData;
    char            iconvCharset[8];
    UbData          uniData;

    /**************************************************************/
    /* The elements in each pointer to structure of the following */
    /* mentioned pointers will be set to NULL automatically as    */
    /* soon as the pointer is allocated, because we use calloc()  */
    /* functions which sets the allocated buffer to NULL values.  */
    /* P O I N T E R S:                                           */
    /*      plh:          plh->methods                            */
    /*                    plh->core.shape_context_size            */
    /*                    plh->core.orientation                   */
    /*                    plh->core.context                       */
    /*                    plh->core.type_of_text                  */
    /*                    plh->core.implicit_alg                  */
    /*                    plh->core.swapping                      */
    /*                    plh->core.numerals                      */
    /*                    plh->core.text_shaping                  */
    /*                                                            */
    /*      defaultData:  defaultData->methods                    */
    /*                                                            */
    /*      arabicData:   arabicData->Word                        */
    /*                    arabicData->OneCell                     */
    /*                    arabicData->SpecialSh                   */
    /**************************************************************/
    plh                  = NULL;
    defaultData          = NULL;
    arabicData           = NULL;
    uniData              = NULL;
    bidivalues           = NULL;

    /*********************************************************************/
    /* allocate the LayoutObject and related structures.                 */
    /*********************************************************************/
    plh = (LayoutObj) calloc(1, sizeof(LayoutObjectRec));
    if (!plh) 
	goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->methods = (LayoutMethods) calloc(1, sizeof(LayoutMethodsRec));
    if (!plh->methods) 
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    defaultData = (Default) calloc(1, sizeof(DefaultRec));
    if(!defaultData)
	goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    arabicData = (Arabic) calloc(1, sizeof(ArabicRec));
    if(!arabicData)
	goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    defaultData->methods = (DefaultMethods) calloc(1, sizeof(DefaultMethodsRec));
    if(!defaultData->methods)
	goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    defaultData->compc = TRUE;
    uniData = (UbData) calloc(1, sizeof(UbDataRec));
    if(!uniData)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;

    arabicData->uni_data = uniData;

    /*********************************************************************/
    /* allocate and set methods                                          */
    /*********************************************************************/
    plh->methods->create       = BidiCreate;
    plh->methods->transform    = BidiTransform;
    plh->methods->wcstransform = wcsBidiTransform;
    plh->methods->setvalues    = BidiSetValues;
    plh->methods->getvalues    = BidiGetValues;
    plh->methods->destroy      = BidiClose;

//    defaultData->methods->arabicEditShape = BidiEditShape;

    /* allocate and set layout values */
    bidivalues = (LayoutCore) calloc(1, sizeof(LayoutCoreRec));
    if (!bidivalues) {
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    }
    /* BiDirection supported in Hebrew and Arabic locales  */
    plh->core.active_dir   = TRUE;
    bidivalues->active_dir = TRUE;

    arabicData->temp_count = 0;
    arabicData->temp_index = 0;

    setlocale(LC_ALL,locale_name); 

    if(SetCurrentCodeSet(bidivalues) == -1)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;

    if((1, 2) == (__LONG_LONG_PAIR(1, 2)))
        /*| Case BIG ENDIAN */
        strcpy(iconvCharset, "UCS-4BE");
    
    else
        /*| Case LITTLE ENDIAN */
        strcpy(iconvCharset, "UCS-4LE");
     
    strcpy(uniData->InCharSet, bidivalues->shape_charset);
    strcpy(uniData->OutCharSet, iconvCharset);
    strcpy(bidivalues->shape_charset, iconvCharset);

    plh->core = *(bidivalues);

    /* check if we need to open iconv */
    if ((j=strcmp(uniData->InCharSet,uniData->OutCharSet))!=0) {
        uniData->iconv_handle_in  = iconv_open(uniData->OutCharSet,uniData->InCharSet);
        uniData->iconv_handle_out = iconv_open(uniData->InCharSet, uniData->OutCharSet); 
    }
    else {
        uniData->iconv_handle_in  = (iconv_t)-1;
        uniData->iconv_handle_out = (iconv_t)-1;
    }

    if ((strcmp(uniData->OutCharSet, "UTF-8") == 0) ||
        (strcmp(uniData->OutCharSet, iconvCharset) == 0) ||
        (strcmp(uniData->InCharSet, ARABIC_CHARSET1) == 0))
        /* Shaping supported in Arabic locales */
        plh->core.active_shape_editing = TRUE;
    /* Shaping not supported in Hebrew */
    else
        plh->core.active_shape_editing = FALSE;

    arabicData->ShapeState = 'I';       /* Initial state */

    /* current code set size */
    plh->core.shape_charset_size = 1;   /* characters to be checked  for
                                         * ShapeEditing */
    plh->core.shape_context_size = (LayoutEditSize) calloc(1, sizeof(LayoutEditSizeRec));
    if(!plh->core.shape_context_size)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;

    /*************************************************************************/
    /* According to statistics, an Arabic character shape may change in      */
    /* correspondance to any change in 3 characters before it and 2 chars    */
    /* after.                                                                */
    /*************************************************************************/
    plh->core.shape_context_size->front = 2;
    plh->core.shape_context_size->back  = 3;

    /* default orientation */
    plh->core.orientation = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.orientation)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.orientation->inp = ORIENTATION_LTR;
    plh->core.orientation->out = ORIENTATION_LTR;

    /* default contextual orientation */
    plh->core.context = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.context)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.context->inp = CONTEXT_LTR;
    plh->core.context->out = CONTEXT_LTR;

    /* default text type */
    plh->core.type_of_text = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.type_of_text)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.type_of_text->inp = TEXT_IMPLICIT;
    plh->core.type_of_text->out = TEXT_VISUAL;

    /* default Bidi Type */
    plh->core.implicit_alg = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.implicit_alg)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.implicit_alg->inp = ALGOR_BASIC;
    plh->core.implicit_alg->out = ALGOR_BASIC;

    /* default Swapping mode */
    plh->core.swapping = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.swapping)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.swapping->inp = SWAPPING_NO;
    plh->core.swapping->out = SWAPPING_YES;

    /* default numeric shapes */
    plh->core.numerals = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.numerals)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    plh->core.numerals->inp = NUMERALS_NOMINAL;
    /*
     * If ShapeEditing is inactive, this means we are Hebrew, so set numerics
     * to nominal 
     */
    if (plh->core.active_shape_editing)
        plh->core.numerals->out = NUMERALS_CONTEXTUAL;
    else
        plh->core.numerals->out = NUMERALS_NOMINAL;

    /* default Shaping mode */
    plh->core.text_shaping = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!plh->core.text_shaping)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR; 
    plh->core.text_shaping->inp = TEXT_NOMINAL;
    /*
     * If ShapeEditing is inactive, this means we are Hebrew, so set shaping
     * to nominal 
     */
    if (plh->core.active_shape_editing)
        plh->core.text_shaping->out = TEXT_SHAPED;
    else
        plh->core.text_shaping->out = TEXT_NOMINAL;

    /* default check mode */
    plh->core.check_mode = MODE_EDIT;

    /**********************************************************************/
    /* Allocate and set private data, these data are needed to handle     */
    /* specific c/c's of the Araibic lang which is :                      */
    /*                                                                    */
    /* Word break   (performing shaping on every single word in isolation */
    /*               from adjacent words)                                 */
    /* OneCell      (to indicate wheather the seen family of chars are to */
    /*               be drawn using one cell or two)                      */
    /* SpecialShape (to indicate wheather the LamAlef char is required to */
    /*              be drawn after shaping on a single cell or two        */
    /**********************************************************************/
    arabicData->Word = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!arabicData->Word)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    arabicData->OneCell = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!arabicData->OneCell)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;
    arabicData->SpecialSh = (LayoutTextDescriptor) calloc(1, sizeof(LayoutTextDescriptorRec));
    if(!arabicData->SpecialSh)
        goto FREE_RESOURCES_ON_ALLOCATION_ERROR;

    /* default word break mode */
    arabicData->Word->inp = NO_BREAK;
    arabicData->Word->out = NO_BREAK;

    /* default onecell mode */
    arabicData->OneCell->inp = ONECELL_SEEN;
    arabicData->OneCell->out = TWOCELL_SEEN;

    /* default AIX/HOST mode */
    arabicData->SpecialSh->inp=TEXT_STANDARD;
    arabicData->SpecialSh->out=TEXT_STANDARD;
        
    defaultData->locale_data=(Arabic)arabicData;
    defaultData->compc = TRUE;
    plh->private_data=(Default)defaultData;

    return (plh);

FREE_RESOURCES_ON_ALLOCATION_ERROR:
    errno = ENOMEM;
    if(plh) {
        if(plh->methods)
            free(plh->methods);
        if(plh->core.shape_context_size)
            free(plh->core.shape_context_size);
        if(plh->core.orientation)
            free(plh->core.orientation);
        if(plh->core.context)
            free(plh->core.context);
        if(plh->core.type_of_text)
            free(plh->core.type_of_text);
        if(plh->core.implicit_alg)
            free(plh->core.implicit_alg);
        if(plh->core.swapping)
            free(plh->core.swapping);
        if(plh->core.numerals)
            free(plh->core.numerals);
        if(plh->core.text_shaping);
            free(plh->core.text_shaping);
        free(plh);
    }
    if(defaultData) {
        if(defaultData->methods)
       	    free(defaultData->methods);
        free(defaultData);
    }
    if(arabicData) {
        if(arabicData->Word)
            free(arabicData->Word); 
        if(arabicData->OneCell)
            free(arabicData->OneCell); 
        if(arabicData->SpecialSh)
            free(arabicData->SpecialSh);
	free(arabicData);
    }
    if(uniData)
	free(uniData);
    if(bidivalues)
        free(bidivalues);
    return(NULL);
}

/***********************************************************************/
/* FUNCTION NAME: BidiCreate()                                         */
/*                                                                     */
/* FUNCTION DESCRIPTION: This routine is used to add the layout values */
/*                       or the BiDi attributes to the newley created  */
/*                       layout object by calling BidiSetvalues()      */
/***********************************************************************/
LayoutObject 
BidiCreate(LayoutObject layoutObject, LayoutValues layoutValues)
{
    int             j = 0;
    char            *cp;
    LayoutCore      bidivalues;
    Default         defaultData;
    Arabic          arabData;

    if (layoutObject->private_data) {
        defaultData = (Default) layoutObject->private_data;
        arabData    = (Arabic) defaultData->locale_data;
    }
    /*
     * A call is needed here to the setvalues function in order for the 
     * modifier to have an effect on the layout object.
     */
    BidiSetValues(layoutObject, layoutValues, &j);
    return (layoutObject);
}
