/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group

******************************************************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "uc_bdstruct.h"
#include "arabic.h"
#include "layout_int.h"
#include "layout_backward.h"

/*********************************************************************/
/* FUNCTION NAME: BidiGetValues()                                    */
/*                                                                   */
/* This is the implementation of the m_getvalues_layout()            */
/* function as defined in the Open Group CAE specification.          */
/*                                                                   */
/* The "layout_object" argument specifies a LayoutObject             */
/* returned by the m_create_layout() function.                       */
/*                                                                   */
/* The "values" argument specifies the list of layout values         */
/* that are to be queried.                                           */
/*                                                                   */
/* If the layout value name has QueryValueSize ORed to it,           */
/* instead of the value of the layout value, only its size is        */
/* returned.                                                         */
/*                                                                   */
/* When the function completes without errors a zero is returned.    */
/* If any value cannot be queried, the index of the value            */
/* causing the error is returned in "index_returned", and a -1       */
/* value is returned.                                                */
/*********************************************************************/
int 
BidiGetValues(LayoutObj context, LayoutValues layout,
              int *index_returned)
{
    int                  result = 0;
    int                  i = 0;
    int                  checkMode;
    LayoutEditSize       ConSize;
    LayoutEditSize       *PConSize;
    LayoutTextDescriptor Descr;
    LayoutTextDescriptor *PDescr;
    BooleanValue         bool;

    Default         defaultData;
    Arabic          arabicData;

    LayoutCore      Values;

    unsigned long   Size;

    if (!layout)
        return (result);

    defaultData = context->private_data;
    arabicData = (Arabic) defaultData->locale_data;
    Values = &context->core;

    if (layout)
        while (layout[i].name != 0) {   /* while not end of array */
            if ((layout[i].name & AllTextDescriptors) &&
                (layout[i].name & QueryValueSize)) {
                layout[i].value = 
                        (unsigned long *) sizeof(LayoutTextDescriptorRec);
            } else if (layout[i].name == AllTextDescriptors) {
                /*************************************************************
                 *
                 * one can query more than one value simultaneously by ORing
                 * the required names and the returned values are ORed together 
                 * this is not in the X/Open specs and was adopted from AIX
                 * old bidi engines.
                 *************************************************************/
                LayoutTextDescriptor temp;
                LayoutTextDescriptor *dummy = NULL;
                int                  dummy_allocated = 0;
                temp = (LayoutTextDescriptor) 
                       calloc(1, sizeof(LayoutTextDescriptorRec));
                if(!temp) {
                    errno = ENOMEM;
                    return(-1);
                }
                if (!layout[i].value) {
                    dummy = (LayoutTextDescriptor *) calloc(1, sizeof(temp));
                    if(!dummy) {
                        errno = ENOMEM;
                        free(temp);
                        return(-1);
                    }
                    dummy_allocated = 1;
                    layout[i].value = (LayoutTextDescriptor *) dummy;
                } else {
                    dummy = (LayoutTextDescriptor *) layout[i].value;
                    dummy_allocated = 0;
                }

                (*dummy)->inp |= context->core.orientation->inp;
                (*dummy)->out |= context->core.orientation->out;

                (*dummy)->inp |= context->core.context->inp;
                (*dummy)->out |= context->core.context->out;

                (*dummy)->inp |= context->core.type_of_text->inp;
                (*dummy)->out |= context->core.type_of_text->out;

                (*dummy)->inp |= context->core.implicit_alg->inp;
                (*dummy)->out |= context->core.implicit_alg->out;

                (*dummy)->inp |= context->core.swapping->inp;
                (*dummy)->out |= context->core.swapping->out;

                (*dummy)->inp |= context->core.numerals->inp;
                (*dummy)->out |= context->core.numerals->out;

                (*dummy)->inp |= context->core.text_shaping->inp;
                (*dummy)->out |= context->core.text_shaping->out;
  
                if(temp) {
                    free(temp);
                    temp = NULL;
                }
                if(dummy && dummy_allocated) {
                    free(dummy);
                    dummy = NULL;
                }
            } else if (layout[i].name & AllTextDescriptors) {

                /*************************************************************
                 *
                 * one can query more than one value simultaneously by ORing
                 * the required names and the returned values are ORed together
                 * this is not in the X/Open specs and was adopted from AIX
                 * old bidi engines.
                *************************************************************/
                LayoutTextDescriptor temp;
                LayoutTextDescriptor *dummy = NULL;
                int                  dummy_allocated = 0;
                temp = (LayoutTextDescriptor) 
                       calloc(1, sizeof(LayoutTextDescriptorRec));
                if(!temp) {
                    errno = ENOMEM;
                    return(-1);
                }
                if (!(layout[i].value)) {
                    dummy = (LayoutTextDescriptor *) calloc(1, sizeof(temp));
                    if(!dummy) {
                        errno = ENOMEM;
                        free(temp);
                        return(-1);
                    }
                    dummy_allocated = 1;
                    layout[i].value = (LayoutTextDescriptor *) dummy;
                } else {
                    dummy = (LayoutTextDescriptor *) layout[i].value;
                    dummy_allocated = 0;
                    (*dummy)->inp = 0;
                    (*dummy)->out = 0;
                }
                if (layout[i].name & Orientation) {
                    (*dummy)->inp |= context->core.orientation->inp;
                    (*dummy)->out |= context->core.orientation->out;
                } if (layout[i].name & Context) {
                    (*dummy)->inp |= context->core.context->inp;
                    (*dummy)->out |= context->core.context->out;
                } if (layout[i].name & TypeOfText) {
                    (*dummy)->inp |= context->core.type_of_text->inp;
                    (*dummy)->out |= context->core.type_of_text->out;
                } if (layout[i].name & ImplicitAlg) {
                    (*dummy)->inp |= context->core.implicit_alg->inp;
                    (*dummy)->out |= context->core.implicit_alg->out;
                } if (layout[i].name & Swapping) {
                    (*dummy)->inp |= context->core.swapping->inp;
                    (*dummy)->out |= context->core.swapping->out;
                } if (layout[i].name & Numerals) {
                    (*dummy)->inp |= context->core.numerals->inp;
                    (*dummy)->out |= context->core.numerals->out;

                } if (layout[i].name & TextShaping) {
                    (*dummy)->inp |= context->core.text_shaping->inp;
                    (*dummy)->out |= context->core.text_shaping->out;
                }
                if(temp) {
                    free(temp);
                    temp = NULL;
                }
                if(dummy && dummy_allocated) {
                    free(dummy);
                    dummy = NULL;
                }
            } else if (layout[i].name & ActiveDirectional) {
                if (layout[i].name & QueryValueSize) {
                    layout[i].value = 
                            (unsigned long *) sizeof(context->core.active_dir);
                } else {
                    BooleanValue   *dummy = layout[i].value;

                    *dummy = (BooleanValue) (context->core.active_dir);
                }
            } else if (layout[i].name & ActiveShapeEditing) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.active_shape_editing);
                } else {
                    BooleanValue   *dummy = layout[i].value;

                    *dummy = context->core.active_shape_editing;
                }
            } else if (layout[i].name & ShapeCharset) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = (unsigned long)
                             strlen(context->core.shape_charset) + 1;
                } else {
                    char **dummy = (char **) layout[i].value;

                    strcpy((*dummy), context->core.shape_charset);
                }
            } else if (layout[i].name & ShapeCharsetSize) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.shape_charset_size);
                } else {
                    int            *dummy = layout[i].value;

                    *dummy = context->core.shape_charset_size;
                }
            } else if (layout[i].name & WordBreak) {
                LayoutTextDescriptor *dummy = (LayoutTextDescriptor *) 
                                              layout[i].value;
                (*dummy)->inp |= arabicData->Word->inp;
                (*dummy)->out |= arabicData->Word->out;

            } else if (layout[i].name & ArabicSpecialShaping) {
                LayoutTextDescriptor *dummy = (LayoutTextDescriptor *) 
                                              layout[i].value;
                (*dummy)->inp |= arabicData->SpecialSh->inp;
                (*dummy)->out |= arabicData->SpecialSh->out;
            } else if (layout[i].name & ArabicOneCellShaping) {
                LayoutTextDescriptor *dummy = (LayoutTextDescriptor *)
                                              layout[i].value;
                (*dummy)->inp |= arabicData->OneCell->inp;
                (*dummy)->out |= arabicData->OneCell->out;
            } else if (layout[i].name & InOutTextDescrMask) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.in_out_text_descr_mask);
                } else {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = context->core.in_out_text_descr_mask;
                }
            } else if (layout[i].name & InOnlyTextDescr) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.in_only_text_descr);
                } else {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = 0;

                    if (context->core.in_out_text_descr_mask & Orientation)
                        *dummy |= context->core.orientation->inp;
                    if (context->core.in_out_text_descr_mask & Context)
                        *dummy |= context->core.context->inp;
                    if (context->core.in_out_text_descr_mask & TypeOfText)
                        *dummy |= context->core.type_of_text->inp;
                    if (context->core.in_out_text_descr_mask & ImplicitAlg)
                        *dummy |= context->core.implicit_alg->inp;
                    if (context->core.in_out_text_descr_mask & Swapping)
                        *dummy |= context->core.swapping->inp;
                    if (context->core.in_out_text_descr_mask & Numerals)
                        *dummy |= context->core.numerals->inp;
                    if (context->core.in_out_text_descr_mask & TextShaping)
                        *dummy |= context->core.text_shaping->inp;
                }
            } else if (layout[i].name & OutOnlyTextDescr) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.out_only_text_descr);
                } else {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = 0;

                    if (context->core.in_out_text_descr_mask & Orientation)
                        *dummy |= context->core.orientation->out;
                    if (context->core.in_out_text_descr_mask & Context)
                        *dummy |= context->core.context->out;
                    if (context->core.in_out_text_descr_mask & TypeOfText)
                        *dummy |= context->core.type_of_text->out;
                    if (context->core.in_out_text_descr_mask & ImplicitAlg)
                        *dummy |= context->core.implicit_alg->out;
                    if (context->core.in_out_text_descr_mask & Swapping)
                        *dummy |= context->core.swapping->out;
                    if (context->core.in_out_text_descr_mask & Numerals)
                        *dummy |= context->core.numerals->out;

                    if (context->core.in_out_text_descr_mask & TextShaping)
                        *dummy |= context->core.text_shaping->out;
                }
            } else if (layout[i].name & CheckMode) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(context->core.check_mode);
                } else {
                    int            *dummy = layout[i].value;

                    *dummy = context->core.check_mode;
                }
            } else if (layout[i].name & ShapeContextSize) {
                if (layout[i].name & QueryValueSize) {
                    unsigned long  *dummy = layout[i].value;

                    *dummy = sizeof(LayoutEditSizeRec);
                } else {
                    LayoutEditSize *dummy = (LayoutEditSize *) layout[i].value;

                    (*dummy)->front = context->core.shape_context_size->front;
                    (*dummy)->back = context->core.shape_context_size->back;
                }

            } else
                break;          /* Error condition */
            i++;
        }
    return (0);
}
