/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <langinfo.h>
#include "../layout/layout.h"
#include "uc_bdstruct.h"
#include "arabic.h"
#include "layout_int.h"
#include "layout_backward.h"
#include "uc_data.h"

/*********************************************************************/
/* This is the implementation of the m_setvalues_layout()            */
/* function as defined in the Open Group CAE specification.          */
/* The "layout_object" argument specifies a LayoutObject             */
/* returned by the m_create_layout() function.                       */
/* The "values" argument specifies the list of layout values         */
/* that are to be changed.                                           */
/* If the layout value name has QueryValueSize ORed to it,           */
/* instead of the value of the layout value, only its size is        */
/* returned.                                                         */
/* More than one value can be set simultabeously by ORing the        */
/* values to be set together.                                        */
/* When the function completes without errors a zero is returned.    */
/* If any value cannot be set, the index of the value                */
/* causing the error is returned in "index_returned", and a -1       */
/* value is returned.                                                */
/*********************************************************************/
int 
BidiSetValues(LayoutObject contex, LayoutValues layout,
              int *index_returned)
{
    int                  i = 0;
    int                  Invalue, Outvalue;
    LayoutTextDescriptor *PDescr, Descr;
    LayoutCore           Values;
    Default              defaultData;
    Arabic               arabicData;
    UbData               uniData;

    defaultData = (Default) contex->private_data;
    arabicData  = (Arabic) defaultData->locale_data;
    uniData     = (UbData) arabicData->uni_data;

    Values      = &(contex->core);
    if(layout) {
        while (layout[i].name != 0) {   /* while not end of array */
            switch (layout[i].name) {
                case ActiveDirectional:
                case ActiveShapeEditing:
                case ShapeCharsetSize:
                case ShapeContextSize:
                /* should not be set */
                    return (-1);
                    break;
                case ShapeCharset:
                    if (Values->active_shape_editing) {
                        char           *bigLanginfo, *bigCharset, ch;
                        int             count;
                        
                        if(Values->shape_charset)
                            free(Values->shape_charset);
                        bigCharset = (char *) 
                                     calloc(strlen(layout[i].value) + 1, sizeof(char));
                        if(!bigCharset)
                        {
                            errno = ENOMEM;
                            return(-1);
                        }
                        strcpy(bigCharset, layout[i].value);
                        for (count = 0; count < strlen(layout[i].value);
                             count++) {
                            ch = toupper(bigCharset[count]);
                            bigCharset[count] = ch;
                        }
                        bigLanginfo = (char *) nl_langinfo(CODESET);
                        if (strstr(bigLanginfo, bigCharset) 
                            || (strstr(ARABIC_CHARSET1, bigCharset) && 
                                strstr(ARABIC_CHARSET2, bigLanginfo))
                            || strstr(ARABIC_CHARSET1, bigCharset)) {
                            if (strcmp(bigLanginfo, ARABIC_CHARSET2) == 0)
                                bigLanginfo = ARABIC_CHARSET1;
                            Values->shape_charset = (char *) 
                                    calloc(strlen(bigLanginfo) + 1, sizeof(char));
                            if(!Values->shape_charset)
                            {
                                errno = ENOMEM;
			        free(bigCharset);
                                return(-1);
                            }
                            strcpy(Values->shape_charset, bigCharset);
                        } else {
                            Values->shape_charset = (char *)
                                    calloc(strlen("UCS-2") + 1, sizeof(char));
                            if(!Values->shape_charset)
                            {
                                errno = ENOMEM;
				free(bigCharset);
                                return(-1);
                            }
                            strcpy(Values->shape_charset, "UCS-2");
                        }
                        Values->shape_charset_size = 1;/* should be derived from
                                                        * charset
                                                        */
                        /* check if we need to open iconv */
                        if (strcmp(Values->shape_charset, uniData->InCharSet)
                            != 0) {
                            strcpy(uniData->InCharSet, Values->shape_charset);
                            iconv_close(uniData->iconv_handle_out);
                            uniData->iconv_handle_out =
                                    iconv_open(uniData->InCharSet,
                                               uniData->OutCharSet);
                        }
                        free(bigCharset);
                    }
                    break;
                default:    /* otherwise it must be text descriptors */
                            /* check valid name */
                    if ((layout[i].name & ~MaskAllTextDescriptors)
                        && (layout[i].name & ~MaskArabic)) {
                        *index_returned = i;
                        return (-1);
                    }
                    /*
                     * now check on layout[i].name to see which values are
                     * needed 
                     */
                    Descr = (LayoutTextDescriptor) layout[i].value;
                    if (layout[i].name & TypeOfText) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskTypeOfText;
                        Outvalue = Descr->out & MaskTypeOfText;
                        /* check if valid */
                        switch (Invalue) {
                            case TEXT_VISUAL:
                            case TEXT_IMPLICIT:
                            case TEXT_EXPLICIT:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case TEXT_VISUAL:
                            case TEXT_IMPLICIT:
                            case TEXT_EXPLICIT:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        Values->type_of_text->inp = Invalue;
                        Values->type_of_text->out = Outvalue;
                        /* check if other values need change */
                        /* V--V and contex num */
                        if ((Values->type_of_text->inp == TEXT_VISUAL)
                            && (Values->type_of_text->out == TEXT_VISUAL)
                            && (Values->numerals->out == NUMERALS_CONTEXTUAL))
                            Values->numerals->out = NUMERALS_NOMINAL;
                        /* implicit with special shapes */
                        if (((Values->type_of_text->inp == TEXT_IMPLICIT)
                            || (Values->type_of_text->out == TEXT_IMPLICIT))
                            && (Values->text_shaping->out != TEXT_SHAPED)
                            && (Values->text_shaping->out != TEXT_NOMINAL))
                            Values->text_shaping->out = TEXT_NOMINAL;
                    }
                    if (layout[i].name & Orientation) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskOrientation;
                        Outvalue = Descr->out & MaskOrientation;
                        /* check if valid */
                        switch (Invalue) {
                            case ORIENTATION_CONTEXTUAL:
                            case ORIENTATION_LTR:
                            case ORIENTATION_RTL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case ORIENTATION_CONTEXTUAL:
                            case ORIENTATION_LTR:
                            case ORIENTATION_RTL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        Values->orientation->inp = Invalue;
                        Values->orientation->out = Outvalue;
                    }
                    if ((layout[i].name & Context) == Context) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskContext;
                        Outvalue = Descr->out & MaskContext;
                        /* check if valid */
                        switch (Invalue) {
                            case CONTEXT_LTR:
                            case CONTEXT_RTL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case CONTEXT_LTR:
                            case CONTEXT_RTL:
                                break;
                        default:
                            *index_returned = i;
                            return (-1);
                            break;
                        }
                        Values->context->inp = Invalue;
                        Values->context->out = Outvalue;
                    }
                    if (layout[i].name & ImplicitAlg) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskImplicitAlg;
                        Outvalue = Descr->out & MaskImplicitAlg;
                        /* check if valid */
                        switch (Invalue) {
                            case ALGOR_BASIC:
                            case ALGOR_IMPLICIT:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                            break;
                        }
                        switch (Outvalue) {
                            case ALGOR_BASIC:
                            case ALGOR_IMPLICIT:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        Values->implicit_alg->inp = Invalue;
                        Values->implicit_alg->out = Outvalue;
                    }
                    if(layout[i].name & Swapping) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskSwapping;
                        Outvalue = Descr->out & MaskSwapping;
                        /* check if valid */
                        switch (Invalue) {
                            case SWAPPING_YES:
                            case SWAPPING_NO:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case SWAPPING_YES:
                            case SWAPPING_NO:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        Values->swapping->inp = Invalue;
                        Values->swapping->out = Outvalue;
                    }
                    if (layout[i].name & Numerals) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskNumerals;
                        Outvalue = Descr->out & MaskNumerals;
                        /* check if valid */
                        switch (Invalue) {
                            case NUMERALS_NOMINAL:
                            case NUMERALS_NATIONAL:
                            case NUMERALS_CONTEXTUAL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case NUMERALS_NOMINAL:
                            case NUMERALS_NATIONAL:
                            case NUMERALS_CONTEXTUAL:
                                break;
                            default:
                                *index_returned = i;
                            return (-1);
                                break;
                        }
                        /* check invalid combination V-->V + contextual */
                        if ((Values->type_of_text->inp == TEXT_VISUAL)
                            && (Values->type_of_text->out == TEXT_VISUAL)
                            && (Outvalue == NUMERALS_CONTEXTUAL)) {
                            *index_returned = i;
                            return (-1);
                        }
                        /* check if special numerics allowed */
                        if (!(contex->core.active_shape_editing)
                            && (Outvalue != NUMERALS_NOMINAL)) {
                           *index_returned = i;
                            return (-1);
                        }
                        /* assign value */
                        Values->numerals->inp = Invalue;
                        Values->numerals->out = Outvalue;
                    }
                    if (layout[i].name & TextShaping) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskTextShaping;
                        Outvalue = Descr->out & MaskTextShaping;
                        /* check if valid */
                        switch (Invalue) {
                            case TEXT_SHAPED:
                            case TEXT_NOMINAL:
                            case TEXT_SHFORM1:
                            case TEXT_SHFORM2:
                            case TEXT_SHFORM3:
                            case TEXT_SHFORM4:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case TEXT_SHAPED:
                            case TEXT_NOMINAL:
                            case TEXT_SHFORM1:
                            case TEXT_SHFORM2:
                            case TEXT_SHFORM3:
                            case TEXT_SHFORM4:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /*
                         * check invalid combination Implicit +special shapes if
                         * (((Values->type_of_text->inp==TEXT_IMPLICIT)
                         * ||(Values->type_of_text->out==TEXT_IMPLICIT))
                         * &&(Outvalue!=TEXT_SHAPED)&&(Outvalue!=TEXT_NOMINAL)){
                         * index_returned=i; return(-1); } 
                         */
                        /* check if shaping allowed */
                        if (!(contex->core.active_shape_editing)
                            && (Outvalue != TEXT_NOMINAL)) {
                            *index_returned = i;
                            return (-1);
                        }
                        /* assign value */
                        Values->text_shaping->inp = Invalue;
                        Values->text_shaping->out = Outvalue;
                    }
                    if (layout[i].name & WordBreak) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskWordBreak;
                        Outvalue = Descr->out & MaskWordBreak;
                        /* check if valid */
                        switch (Invalue) {
                            case (int)BREAK:
                            case NO_BREAK:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case (int)BREAK:
                            case NO_BREAK:
                               break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        arabicData->Word->inp = Invalue;
                        arabicData->Word->out = Outvalue;
                    }
                    if (layout[i].name & ArabicSpecialShaping) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskSpecialShaping;
                        Outvalue = Descr->out & MaskSpecialShaping;
                        /* check if valid */
                        switch (Invalue) {
                            case TEXT_STANDARD:
                            case TEXT_COMPOSED:
                            case TEXT_SPECIAL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case TEXT_STANDARD:
                            case TEXT_COMPOSED:
                            case TEXT_SPECIAL:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        arabicData->SpecialSh->inp = Invalue;
                        arabicData->SpecialSh->out = Outvalue;
                    }
                    if (layout[i].name & ArabicOneCellShaping) {
                        /* extract needed value */
                        Invalue = Descr->inp & MaskOneCellShaping;
                        Outvalue = Descr->out & MaskOneCellShaping;
                        /* check if valid */
                        switch (Invalue) {
                            case ONECELL_SEEN:
                            case TWOCELL_SEEN:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        switch (Outvalue) {
                            case ONECELL_SEEN:
                            case TWOCELL_SEEN:
                                break;
                            default:
                                *index_returned = i;
                                return (-1);
                                break;
                        }
                        /* assign value */
                        arabicData->OneCell->inp = Invalue;
                        arabicData->OneCell->out = Outvalue;
                    }
                    if (layout[i].name & InOutTextDescrMask) {
                        contex->core.in_out_text_descr_mask = (unsigned long)
                                                        layout[i].value;
                    } else if (layout[i].name & InOnlyTextDescr) {
                        unsigned long dummy = (unsigned long)layout[i].value;
                     
                        if (contex->core.in_out_text_descr_mask & Orientation)
                            contex->core.orientation->inp=
                                    (dummy & MaskOrientation);
                        if (contex->core.in_out_text_descr_mask & Context)
                            contex->core.context->inp = (dummy & MaskContext);
                        if (contex->core.in_out_text_descr_mask & TypeOfText)
                            contex->core.type_of_text->inp =
                                    (dummy & MaskTypeOfText);
                        if (contex->core.in_out_text_descr_mask & ImplicitAlg)
                           contex->core.implicit_alg->inp =
                                    (dummy & MaskImplicitAlg);
                        if (contex->core.in_out_text_descr_mask & Swapping)
                            contex->core.swapping->inp = (dummy & MaskSwapping);
                        if (contex->core.in_out_text_descr_mask & Numerals)
                            contex->core.numerals->inp = (dummy & MaskNumerals);
                        if (contex->core.in_out_text_descr_mask & TextShaping)
                           contex->core.text_shaping->inp =
                                   (dummy & MaskTextShaping);
                    } else if (layout[i].name & OutOnlyTextDescr) {
                        unsigned long dummy = (unsigned long)layout[i].value;

                        if (contex->core.in_out_text_descr_mask & Orientation)
                            contex->core.orientation->out =
                                    (dummy & MaskOrientation);
                        if (contex->core.in_out_text_descr_mask & Context)
                            contex->core.context->out = (dummy & MaskContext);
                        if (contex->core.in_out_text_descr_mask & TypeOfText)
                            contex->core.type_of_text->out =
                                    (dummy & MaskTypeOfText);
                        if (contex->core.in_out_text_descr_mask & ImplicitAlg)
                            contex->core.implicit_alg->out =
                                    (dummy&MaskImplicitAlg);
                        if (contex->core.in_out_text_descr_mask & Swapping)
                            contex->core.swapping->out = (dummy & MaskSwapping);
                        if (contex->core.in_out_text_descr_mask & Numerals)
                            contex->core.numerals->out = (dummy & MaskNumerals);
                        if (contex->core.in_out_text_descr_mask & TextShaping)
                            contex->core.text_shaping->out =
                                    (dummy&MaskTextShaping);
                    } else if (layout[i].name & CheckMode) {
                        contex->core.check_mode = (int) layout[i].value;
                    }
                    break;
            }
            i++;
        }
    }
    return (0);
}
