/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdlib.h>
#include <errno.h>
#include "uc_bdstruct.h"
#include "layout_int.h"
#include "thai_type.h"

int
ThaiTransform(
              PICSPARAMRec IcsRec,
              const unsigned short * InpBuf,
              const size_t InpSize,
              void *OutBuf,
              size_t * Outsize,
              size_t * InpToOut,
              size_t * OutToInp,
              unsigned char *Property
)
{
    /*
     * Internal State 
     *
     * Sequence : determine the possibility of Composing or Kerning 0
     * one of composing characters allowed CONSONANT all composing 
     * characters allowed CONSONANT + UV or LV only TM/UD1/UD2 
     * characters allowed CONSONANT + TM only VUA character allowed 
     */
    static th_char_t Sequence        = 0;
    size_t           *InpBufIndex, index;
    unsigned short   *inpBuf         = NULL;
    int              i;
    int              sara_am_advance = 0;

    /*
     * Type : determine character type including "New Cell" character 0
     * lternate glyph not necessary 
     */
    th_char_t       Type;

    unsigned short  *Buf = NULL, wch;
    size_t          iInpSize, iIn, iOut, iOutToInp;
    int             iNumShape, iRet = 0;
    LayoutDesc      TextShape;
    LayoutCore      Values;

    index = 0;
    InpBufIndex = &index;
    /* Reset internal state */
    if (InpBuf == NULL) {
        Sequence = 0;
        return 0;
    }
    /* Determine the length of a text delimited by a NULL code element. */
    if (InpSize == -1)
        for (iInpSize = 0; InpBuf[iInpSize] != (unsigned short) NULL; iInpSize++);
    else
        iInpSize = InpSize;

    inpBuf = (unsigned short *) calloc(InpSize, sizeof(unsigned short));
    if(!inpBuf)
    {
        errno = ENOMEM;
        return(-1);
    }
    for (i = 0; i < InpSize; i++)
        inpBuf[i] = InpBuf[i];

    /* Calculate the size of an output buffer required for a transformed text */
    if (*Outsize == 0) {
        *Outsize = iInpSize * MAX_TRANSFORMED_CHAR;
        free(inpBuf);
        if(Buf)
            free(Buf);
        return 0;
    }
    if (IcsRec->num_flag == NUMERALS_NATIONAL)
        iNumShape = THAI_NUM_SHAPE;
    else if (IcsRec->num_flag == NUMERALS_NOMINAL)
        iNumShape = ARABIC_NUM_SHAPE;
    else
        iNumShape = 0;

    /* --------------------------- */
    /* T R A N S F O R M A T I O N */
    /* --------------------------- */

    /* Dummy buffer necessary for transformation */
    if (OutBuf != NULL)
        Buf = (unsigned short *) OutBuf;
    else
    {
        Buf = (unsigned short *)
              calloc(InpSize * MAX_TRANSFORMED_CHAR, sizeof(unsigned short));
        if(!Buf)
        {
            errno = ENOMEM;
            free(inpBuf);
            return(-1);
        }
    }

    TextShape = IcsRec->txtShp_flag;

    for (iIn = iOut = 0; iIn < iInpSize; iIn++, iOut++) {
        /* !!! Output buffer full !!! */
        if (OutBuf && iOut + MAX_TRANSFORMED_CHAR > *Outsize) {
            errno = E2BIG;
            iRet = -1;
            break;
        }
        iOutToInp = iOut;
        wch = inpBuf[*InpBufIndex + iIn];

        if (IS_THAI_UC(wch))
            Type = TYPE(wch);
        else if (IS_ARABIC_NUM(wch))
            Type = NUM;
        else
            Type = BS;

        /* TEXT_SHFORM1: all composite sequence is invalid */
        if (TextShape == TEXT_SHFORM1)
            Sequence = 0;

        /*
         * Exception Handling for an invalid composite sequence - for any
         * composing character: 1) a white space character will be inserted
         * 2) its type will be reset 3) now it can be handled normally 
         */
        if (Type & COMPOSING && ( !Sequence || Sequence & (TM | VRD) || Sequence & NBV && (Type & NBV
                                  || Type & UD2 && !(Sequence & UV3) || Type & UD1 && !(Sequence & (LV1 | UV1))))) {
            Buf[iOut++] = W_SPACE;
            Type = 0;
        }
        switch (Type) {

        case TM:
        case UD1:
        case UD2:

            if (Sequence & CNA)
                Buf[iOut] = wch + ((Sequence & UV) ? 0 : CNA_TM_OFFSET);
            else if (Sequence & CRA)
                Buf[iOut] = wch + ((Sequence & UV) ? CRA_UV_TM_OFFSET
                                   : CRA_TM_OFFSET);
            else if (Sequence & CA)
                Buf[iOut] = wch;
            else;               /* ----- ERROR ----- */

            Sequence = (Sequence & NBV || Type & (UD1 | UD2))
                ? 0 : Sequence | TM;
            break;

        case UV1:
        case UV2:
        case UV3:
        case UD3:

            if (Sequence & NBV) {
                Buf[iOut++] = W_SPACE;
            } else if (Sequence & CNA)
                Buf[iOut] = wch;
            else if (Sequence & CRA)
                Buf[iOut] = wch + CRA_UV_OFFSET;
            else if (Sequence & CA) {
                Buf[iOut - 1] += CA_UV_OFFSET;
                Buf[iOut] = wch;
            } else;             /* ----- ERROR ----- */

            Sequence = (Type & UD3) ? 0 : (Sequence | Type);
            break;

        case LV1:
        case LV2:
        case LD:

            if (Sequence & CND)
                Buf[iOut] = wch;
            else if (Sequence & CD)
                Buf[iOut] = wch + CD_LV_OFFSET;
            else if (Sequence & CUD) {
                Buf[iOut - 1] += CUD_LV_OFFSET;
                Buf[iOut] = wch;
            } else;             /* ----- ERROR ----- */

            Sequence = (Type & LD) ? 0 : (Sequence | Type);
            break;

        case VLA:               /* Base-line Vowel with Left-hanging Ascender
                                 * type */

            if (!Sequence || Sequence & (UV | TM | CRA | CA))
                Buf[iOut] = wch;
            else
                Buf[iOut] = wch + NOASC_VLA_OFFSET;

            Sequence = 0;
            break;

        case VUA:               /* Base-line Vowel with Unconnected Ascender
                                 * type */

            if (!Sequence || Sequence & (NBV | VRD))
                Buf[iOut] = wch;
            else if (Sequence & CNA) {
                if (Sequence & TM)
                    Buf[iOut - 1] -= CNA_TM_OFFSET;

                Buf[iOut++] = NIKKHAHIT + CNA_TM_OFFSET;
                Buf[iOut] = SARA_AA;
                sara_am_advance++;
            } else if (Sequence & CRA) {
                if (Sequence & TM)
                    Buf[iOut - 1] -= CRA_TM_OFFSET - CRA_UV_TM_OFFSET;

                Buf[iOut++] = NIKKHAHIT + CRA_TM_OFFSET;
                Buf[iOut] = SARA_AA;
                sara_am_advance++;
            } else if (Sequence & CA) {
                if (Sequence & TM) {
                    Buf[iOut - 2] += CA_UV_OFFSET;
                    Buf[iOut++] = NIKKHAHIT + CNA_TM_OFFSET;
                } else
                    Buf[iOut++] = NIKKHAHIT;

                Buf[iOut] = SARA_AA;
                sara_am_advance++;
            } else;             /* ----- ERROR ----- */

            Sequence = 0;
            break;

        case NUM:               /* NUMeric type */

            if (iNumShape & THAI_NUM_SHAPE && IS_ARABIC_NUM(wch))
                Buf[iOut] = wch + ARABIC_THAI_DIFF;
            else if (iNumShape & ARABIC_NUM_SHAPE && IS_THAI_NUM(wch))
                Buf[iOut] = wch - ARABIC_THAI_DIFF;
            else
                Buf[iOut] = wch;

            Sequence = 0;
            break;

        default:
            /*
             * CONSONANT - All consonant type VRD - Base-line vowels with right descender BS - Other base-line 
	     * symbol type 0 - Type-reset composing character type 
             */

            /* VRD + SARA_AA */
            if (wch == SARA_AA && Sequence & VRD)
                Buf[iOut] = LAAK_KANG;
            else
                Buf[iOut] = wch;

            Sequence = (Type & (CONSONANT | VRD)) ? Type : 0;
            break;
        }

        if (InpToOut)
            InpToOut[iIn] = iOutToInp + sara_am_advance;
        sara_am_advance = 0;

        if (OutToInp)
            while (iOutToInp <= iOut)
                OutToInp[iOutToInp++] = iIn;

        if (Property)
            Property[iIn] = (Type & COMPOSING) ? 0 : NEW_CELL;
    }

    *Outsize = iOut;            /* Actual length of the transformed text */
    *InpBufIndex += iIn;

    if (iRet != -1 || errno != E2BIG)
        Sequence = 0;
    free(inpBuf);
    if(Buf)
        free(Buf);
    return iRet;
}
