/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <langinfo.h>
#include <iconv.h>
#include <errno.h>
#include "uc_bdstruct.h"
#include "layout_int.h"
#include "layout_backward.h"
#include "arabic.h"
#include "uc_data.h"
#include "csd.h"

#define MULTIPLIER 2


/******************************************************************/
/* FUNCTION NAME: UCharInver()                                    */
/*                                                                */
/* This function inverts a wide character text buffer,            */
/* performing symmetric swapping if required                      */
/******************************************************************/
int UCharInver(PUNI_CHAR buffer, size_t buffer_size, int SymSwap)
{
   UNI_CHAR temp;
   int      i, j;

   if (SymSwap) {
      for (i=0, j=buffer_size-1; i<j; i++, j--) {
         temp=UCQSYMM(buffer[i]);
         buffer[i]=UCQSYMM(buffer[j]);
         buffer[j]=temp;
      }
   }
   else {
      for (i=0, j=buffer_size-1; i<j; i++, j--) {
         temp=buffer[i];
         buffer[i]=buffer[j];
         buffer[j]=temp;
      }
   }

   return 0;
}

/***************************************************************************/
/* FUNTCION NAME: Invertbuffer()                                           */
/*                                                                         */
/* This function inverts a buffer of unsigned int, between lower_limit and */
/* upper_limit .                                                           */
/* (Used to invert InpToOut and OutToInp Mappings)                         */
/***************************************************************************/
int Invertbuffer(unsigned int *buffer, size_t lower_limit, size_t upper_limit)
{
   unsigned int temp;
   int      i,j;

   for (i=lower_limit, j=upper_limit; i<j; i++, j--)
   {
      temp=buffer[i];
      buffer[i]=buffer[j];
      buffer[j]=temp;
   }

   return 0;
}


/***************************************************************************/
/* FUNTCION NAME: BidiTransform()                                          */
/*                                                                         */
/* This function is the one called from the m_transform_layout()           */
/* interface of the X/Open PLS APIs                                        */
/* It is responsible for the transformation of single byte code sets.      */
/***************************************************************************/
int 
BidiTransform(LayoutObject        context,
              unsigned char       *InpBuf,
              size_t              InpSize,
              void                *OutBuf,
              size_t              *OutSize,
              size_t              *ToOutBuf,
              size_t              *ToInpBuf,
              unsigned char       *BidiLvl,
              size_t              *InpBufIndex)
{
    size_t                  *InpToOut    = NULL;
    size_t                  *OutToInp    = NULL;
    size_t                  inpSize      = InpSize;
    size_t                  innum        = 0;
    size_t                  outnum       = 0;
    int                     txtShp_flag  = 0;
    int                     i, x, buflen = 0;
    int                     RC = 0, rc   = 0;
    int                     counter      = 0;
    int                     uc_flag      = 0;
    int                     index        = 0;
    int                     outputIndex  = 0;
    int                     needToInvert = FALSE;
    int                     num_flag;
    int                     strongChar;

    ICSPARAMRec             IcsRec;
    LayoutCore              Values;
    unsigned char           *inpconvbuf  = NULL;
    unsigned char           *outconvbuf  = NULL;
    unsigned char           *outpbuf     = NULL;
    unsigned char           *store       = NULL; 
    unsigned char           *euroSignChecker;
    wchar_t                 *tempbuf     = NULL;
    wchar_t                 *temptrg     = NULL;

    Default                 defaultData;
    Arabic                  arabicData;
    UbData                  uniData;

    IcsRec.orient_in = IcsRec.orient_out = 0;
    Values           = &context->core;
    defaultData      = context->private_data;
    arabicData       = defaultData->locale_data;
    uniData          = arabicData->uni_data;

    if (strcmp(Values->shape_charset, "UTF-8")==0 || strcmp(Values->shape_charset, "UNIVERSAL")==0) 
        uc_flag = 1;
    else
        uc_flag = 0;
    /********************* Start of String Conversion *************** 
     * Determination of the length srcLen of source text to process * 
     ****************************************************************/
    if(!InpBuf || !(*InpBuf)) {
        errno = EINVAL;
        return(-1);
    }
    if(!OutBuf) {
        errno = EINVAL;
        return(-1);
    } 
    if (OutSize == NULL || InpSize == 0) {
        errno = EINVAL;
        RC = -1;
        return (RC);
    }
    /* Decide on length of input string if not given */
    if (inpSize == (size_t) - 1 || inpSize > 10000) {
        i = 0;
        while (InpBuf[i] > (size_t) NULL) {
            if (i >= *OutSize && *OutSize) {
                RC = RC | 1;
                break;
            }
            i++;
        }
        inpSize = i;
        InpSize = i;
    }
    if (*OutSize == 0) { /* calculate size of output buffer and return */
        *OutSize = inpSize;
    }
    if (InpSize > *OutSize) {
        InpSize = *OutSize;
        RC = RC | 1;
        errno = E2BIG;
    }

    if (inpSize > *OutSize) {
        inpSize = *OutSize;
        RC = RC | 1;
    }
    store = (unsigned char *) calloc((int) inpSize, sizeof(unsigned char));
    if(!store) {
        errno = ENOMEM;
        return(-1);
    }
    /* decide on numerical shape flag */
    if (Values->numerals->inp != Values->numerals->out)
        num_flag = Values->numerals->out;
    else
        num_flag = NUMERALS_NONE;
    /********************************************************************/
    /* Copy single-byte input string to wide-character temporary buffer */
    /* to be used in the call to UCB2VIS() function                     */
    /********************************************************************/
    tempbuf = (wchar_t *) calloc((MULTIPLIER * inpSize), sizeof(wchar_t));
    if(!tempbuf) {
        free(store);
        store = (char *)NULL;
        errno = ENOMEM;
        return(-1);
    }
    inpconvbuf = InpBuf;
    outconvbuf = (unsigned char*)tempbuf;
    innum      = inpSize;
    outnum     = inpSize * MULTIPLIER * sizeof(wchar_t);
   
    /* A call to iconv is needed now to convert input buffer from single byte to Unicode */
    /* Ahmad --- The following condition is special to handle the -1 problem of iconv */
    euroSignChecker = inpconvbuf;
    while((euroSignChecker = strchr(euroSignChecker, 0xff)))
        *euroSignChecker++ = 0xfe;
    iconv(uniData->iconv_handle_in, (char **) &inpconvbuf, &innum,(char **) &outconvbuf, &outnum);
    buflen = inpSize * MULTIPLIER * sizeof(wchar_t) - outnum;
    strongChar = StrongChar(tempbuf, ((size_t) buflen / sizeof(wchar_t)));

    /*------------------------------------*/
    /* Decide on intial shaping direction */
    /*------------------------------------*/
    if (Values->type_of_text->inp == TEXT_IMPLICIT) {
        /* for implicit source strings, shaping is always RTL */
        if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_in = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->inp == CONTEXT_LTR) {
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (Values->context->inp == CONTEXT_RTL) {
                IcsRec.orient_in = ORIENTATION_RTL;
            } else {
                IcsRec.orient_in = ORIENTATION_LTR;
            }
        } else
            IcsRec.orient_in = Values->orientation->inp;
        if ((Values->orientation->out == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_out = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_out = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->out == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->out == CONTEXT_LTR) {
                IcsRec.orient_out = ORIENTATION_LTR;
            } else if (Values->context->out == CONTEXT_RTL) {
                IcsRec.orient_out = ORIENTATION_RTL;
            } else {
                IcsRec.orient_out = ORIENTATION_LTR;
            }
        } else {
            IcsRec.orient_in  = Values->orientation->inp;
            IcsRec.orient_out = Values->orientation->out;
        }
    } else {  /* for visual source string, shaping is according to string orient */
        if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_in = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->inp == CONTEXT_LTR) {
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (Values->context->inp == CONTEXT_RTL) {
                IcsRec.orient_in = ORIENTATION_RTL;
            } else {
                IcsRec.orient_in = ORIENTATION_LTR;
            }
        } else {
            IcsRec.orient_in  = Values->orientation->inp;
            IcsRec.orient_out = Values->orientation->out;
        }
    }
    /*-----------------------------------------------------------*/
    temptrg = calloc((inpSize * MULTIPLIER * sizeof(wchar_t)), sizeof(wchar_t));
    if(!temptrg) {
        free(store);
        store = (char *)NULL;
        free(tempbuf);
        tempbuf = (wchar_t *)NULL;
        errno = ENOMEM;
        return(-1);
    }

    InpToOut = calloc((inpSize * MULTIPLIER * sizeof(wchar_t)), sizeof(size_t));
    if (!InpToOut) {
        free(store);
        free(tempbuf);
        free(temptrg);
        errno = ENOMEM;
        return(-1);
    }

    OutToInp = calloc((inpSize * MULTIPLIER * sizeof(wchar_t)), sizeof(size_t));
    if (!OutToInp) {
        free(store);
        free(tempbuf);
        free(temptrg);
        free(InpToOut);
        errno = ENOMEM;
        return(-1);
    }

    for (i = 0; i < (inpSize * MULTIPLIER * sizeof(wchar_t)); i++)
        temptrg[i] = 0;
    for (i = 0; i < (inpSize * MULTIPLIER * sizeof(wchar_t)); i++)
        InpToOut[i] = 0;
    for (i = 0; i < (inpSize * MULTIPLIER * sizeof(wchar_t)); i++)
        OutToInp[i] = 0;
    /*-------------------------------------------------*/
    /* The following fills the IcsRec struct variables */
    /*-------------------------------------------------*/
    index = 0;
    if (Values->type_of_text->inp != Values->type_of_text->out){
        needToInvert = (Values->type_of_text->inp == TEXT_VISUAL)&&
                       (IcsRec.orient_out != IcsRec.orient_in);
 
            IcsRec.order_flag = TRUE;
            if(arabicData->SpecialSh->out == TEXT_SPECIAL){
                IcsRec.compc    = TRUE;
                IcsRec.backward = TRUE;
            } else {
                IcsRec.compc    = defaultData->compc;
                IcsRec.backward = FALSE;
            }

            IcsRec.wordbreak = arabicData->Word->out;
            IcsRec.codeset   = Values->shape_charset;
            IcsRec.num_flag  = num_flag;
            if (Values->text_shaping->inp != Values->text_shaping->out)
                IcsRec.txtShp_flag = Values->text_shaping->out;
            else
                IcsRec.txtShp_flag = FALSE;
            IcsRec.formShp  = 0;
            IcsRec.boc_flag = 0;
            if (Values->type_of_text->inp != Values->type_of_text->out)
                IcsRec.flip_flag = TRUE;
            else
                IcsRec.flip_flag = FALSE;
            IcsRec.symmetric = (Values->swapping->inp!=Values->swapping->out);

            if(InpBufIndex)
                IcsRec.size = (buflen/sizeof(wchar_t)) - *InpBufIndex;
            else
                IcsRec.size = buflen/sizeof(wchar_t);                                              

            IcsRec.buffer_in  = (PUNI_CHAR) & tempbuf[index];
            IcsRec.buffer_out = (PUNI_CHAR) & temptrg[outputIndex];
            if (Values->type_of_text->out == TEXT_VISUAL)
                IcsRec.Tables = TEXT_VISUAL;
            else
                IcsRec.Tables = TEXT_IMPLICIT;
            IcsRec.A_level = &store[index];
 
            IcsRec.SrcToTrgMap = (size_t *)&InpToOut[index];
            IcsRec.TrgToSrcMap = (size_t *)&OutToInp[outputIndex];

            if(needToInvert){
                IcsRec.orient_in = IcsRec.orient_out;
                if( (inverse((char *)IcsRec.buffer_in, 
                     IcsRec.size,sizeof(wchar_t))) == -1) {
                    if(errno == ENOMEM) {
                        free(store);
                        free(tempbuf);
                        free(temptrg);
                        free(InpToOut);
                        free(OutToInp);
                        return(-1);
                    }
                }
            }
 
	    /* Now call the actual transformation process from the function UCB2VIS */
	    /* which will call from within the 5 passes from pass1 to pass5         */ 
            rc = UCB2VIS(&IcsRec);
            if(rc == -1 && errno == ENOMEM) {
                free(store);
                free(tempbuf);
                free(temptrg);
                free(InpToOut);
                free(OutToInp);
                return(-1);
            }
            
            if(needToInvert){
                if(IcsRec.SrcToTrgMap)
                    Invertbuffer(IcsRec.SrcToTrgMap,0,IcsRec.size-1);
                if(IcsRec.TrgToSrcMap)
                    Invertbuffer(IcsRec.TrgToSrcMap,0,IcsRec.size-1);
            } 
           
            RC = rc | RC;
      } else {
            IcsRec.size        = buflen/sizeof(wchar_t);
            IcsRec.buffer_in   = (PUNI_CHAR) & tempbuf[index];
            IcsRec.buffer_out  = (PUNI_CHAR) & tempbuf[outputIndex];
            IcsRec.SrcToTrgMap = &InpToOut[index];
            IcsRec.TrgToSrcMap = &OutToInp[outputIndex];
  
            for (i = 0; i < IcsRec.size; i++)
                IcsRec.TrgToSrcMap[i] = i;
            for (i = 0; i < IcsRec.size; i++)
                IcsRec.SrcToTrgMap[i] = i;
            if (BidiLvl)
                for (i = 0; i < IcsRec.size; i++)
                    BidiLvl[i] = 0x80;
            IcsRec.orient_out = Values->orientation->out;
            IcsRec.orient_in  = Values->orientation->inp; 
 
            needToInvert = (Values->type_of_text->inp == TEXT_VISUAL)&&
                           (IcsRec.orient_out != IcsRec.orient_in);
            if(needToInvert) {
                /* Inversion is required */
                if((inverse(IcsRec.buffer_out,IcsRec.size, sizeof(wchar_t))) == -1) {
                    if(errno = ENOMEM) {
                        free(store);
                        free(tempbuf);
                        free(temptrg);
                        free(InpToOut);
                        free(OutToInp);
                        return(-1);
                    }
                }
                if (ToInpBuf)
                    if((inverse(IcsRec.TrgToSrcMap,
                       IcsRec.size, sizeof(size_t))) == -1) {
                        if(errno = ENOMEM) {
                            free(store);
                            free(tempbuf);
                            free(temptrg);
                            free(InpToOut);
                            free(OutToInp);
                            return(-1);
			}
                    }
                if (ToOutBuf)
                    for (i = 0; i < inpSize; i++)
                        IcsRec.SrcToTrgMap[i] = 
                                inpSize - 1 - IcsRec.SrcToTrgMap[i];
            } else {
                if(ToInpBuf)
                  for (i = 0; i < InpSize; i++)
                     ToInpBuf[i] = i;
                if(ToOutBuf)
                  for (i = 0; i < InpSize; i++)
                     ToOutBuf[i] = i;
            }
        }
    /**********************************************************************/
    if (BidiLvl
        && (Values->type_of_text->inp != Values->type_of_text->out)){
        for (i = 0; i < inpSize; i++) { 
            BidiLvl[i] = IcsRec.A_level[i]; 
        }
    }

    for (i = 0; i < IcsRec.size; i++)
        IcsRec.TrgToSrcMap[i] = IcsRec.TrgToSrcMap[i] & 0x03FF;
    free(store);
    store = NULL;

    /***********************************************************************/
    /*----------------------------------------*/
    /* copy temporary buffer to output buffer */
    /*----------------------------------------*/
    if (OutBuf) {
        inpconvbuf = (unsigned char *) IcsRec.buffer_out;
        outpbuf = (unsigned char *)OutBuf;
        innum = buflen;
        outnum = inpSize * MULTIPLIER * sizeof(wchar_t);
        
        /* Now we'll need to call iconv to convert the buffer from Unicode to the output code page */
        iconv(uniData->iconv_handle_out, (char **) &inpconvbuf,&innum, (char **) &outpbuf, &outnum);
        *OutSize = inpSize * MULTIPLIER * sizeof(wchar_t) - outnum;
        outpbuf = (unsigned char *)OutBuf;

        if ((Values->text_shaping->inp != Values->text_shaping->out)
            && context->core.active_shape_editing
            && !((Values->type_of_text->inp == TEXT_IMPLICIT) || 
                 (Values->type_of_text->out == TEXT_IMPLICIT))) {

            BidiShape(outpbuf, inpSize, arabicData->OneCell->out,
                      arabicData->SpecialSh->out, 1,
                      Values->text_shaping->out);
        }

        /* To Handle the case of single byte LamAlef */
        if( ( (Values->type_of_text->inp == TEXT_IMPLICIT) || (Values->type_of_text->out == TEXT_IMPLICIT) ) &&
            ((i = strcmp(uniData->InCharSet, ARABIC_CHARSET1))==0 || (i = strcmp(uniData->InCharSet, ARABIC_CHARSET2))==0)){   
            unsigned char   *try   = NULL;
            unsigned char   *final = NULL;
            int             step, iend, x;
            if (Values->orientation->out == ORIENTATION_LTR) {
                i = 0;
                iend = inpSize - 1;
                step = 1;
            } else {
                i = inpSize - 1;
                iend = 0;
                step = -1;
            }
            try = OutBuf;
            for (x = i; x != iend; x += step) {
                if ((try[x] == 0xdc || try[x] == 0xdd 
                    || try[x] == 0xde || try[x] == 0x80) 
                    && (try[x + step] == 0xf5)) {
                    switch (try[x]) {
                    case 0xdc:
                        try[x] = 0xa1;
                        break;
                    case 0xdd:
                        try[x] = 0xa2;
                        break;
                    case 0x80:
                        try[x] = 0xa3;
                        break;
                    case 0xde:
                        try[x] = 0xa5;
                        break;
                    }
                }
            }
        }
        /********************* Target To Source Mappings *********************/
        if (ToInpBuf) {
            for (i = 0; i < *OutSize; i++)
                ToInpBuf[i] = 0;
            i = 0;
            x = 0;
            while (i < *OutSize) {
                if ((outpbuf[i] >> 7) == 0 || uc_flag == 0) {
                    ToInpBuf[i] = IcsRec.TrgToSrcMap[x];
                    i++;
                    x++;
                } else if ((outpbuf[i] >> 5) == 0x06) {
                    ToInpBuf[i] = IcsRec.TrgToSrcMap[x];
                    ToInpBuf[i + 1] = IcsRec.TrgToSrcMap[x];
                    i = i + 2;
                    x++;
                } else if ((outpbuf[i] >> 4) == 0x0E) {
                    ToInpBuf[i] = IcsRec.TrgToSrcMap[x];
                    ToInpBuf[i + 1] = IcsRec.TrgToSrcMap[x];
                    ToInpBuf[i + 2] = IcsRec.TrgToSrcMap[x];
                    i = i + 3;
                    x++;
                } else {
                    ;
                }
            }
        }
    }
    /********************* Source To Target Mappings *********************/
    if (ToOutBuf) {
        for (i = 0; i < inpSize; i++)
            ToOutBuf[i] = 0;
        i = 0;
        x = 0;
        while (i < inpSize) {
            if ((InpBuf[i] >> 7) == 0 || uc_flag == 0) {
                ToOutBuf[i] = IcsRec.SrcToTrgMap[x];
                i++;
                x++;
            } else if ((InpBuf[i] >> 5) == 0x06) {
                ToOutBuf[i] = IcsRec.SrcToTrgMap[x];
                ToOutBuf[i + 1] = IcsRec.SrcToTrgMap[x];
                i = i + 2;
                x++;
            } else if ((InpBuf[i] >> 4) == 0x0E) {
                ToOutBuf[i] = IcsRec.SrcToTrgMap[x];
                ToOutBuf[i + 1] = IcsRec.SrcToTrgMap[x];
                ToOutBuf[i + 2] = IcsRec.SrcToTrgMap[x];
                i = i + 3;
                x++;
            } else {
                i++;
            }
        }
    }
   /********************* End of String Conversion *************************/
   free(tempbuf);  /* No check before free because the check was at        */ 
   free(temptrg);  /* allocation time.                                     */
   free(InpToOut); 
   free(OutToInp); 
   return (RC);  /* Successfully performed */
}

/***************************************************************************/
/* FUNTCION NAME: StrongChar()                                             */
/*                                                                         */
/* This function is used to determine if the character is a strong LTR     */
/* character or a strong RTL character.                                    */
/***************************************************************************/                                                      
int 
StrongChar(wchar_t * inputBuffer, size_t length)
{
    UBA_TYPE        ch;
    size_t          i;
    for (i = 0; i < length; i++) {
        ch = getchtype(inputBuffer[i]);

        if (ch == UBAT_L)
            return (1);
        else if (ch == UBAT_R)
            return (2);
    }
    return (0);
}

/***************************************************************************/
/* FUNTCION NAME: inverse()                                                */
/*                                                                         */
/* Inverses a stream of elements, where each elemnt is of size element_size*/
/***************************************************************************/                                                      
int 
inverse(buffer, num_of_elements, element_size)
    unsigned char   *buffer;
    size_t          num_of_elements;
    int             element_size;
{
    unsigned char   *temp;
    int             i, j;
    int             start, end;

    temp = (unsigned char *) calloc(element_size, sizeof(char));
    if(!temp) {
        errno = ENOMEM;
        return(-1);
    }
    memset(temp, '\0', element_size);
    for (i = 0, j = num_of_elements - 1; i < j; i++, j--) {
        memcpy(temp, &(buffer[i * element_size]), element_size);
        memcpy(&(buffer[i * element_size]), 
               &(buffer[j * element_size]), element_size);
        memcpy(&(buffer[j * element_size]), temp, element_size);
    }
    free(temp);
    return(0);
}

/***************************************************************************/
/* FUNTCION NAME: BidiShape()                                              */
/*                                                                         */
/* This is the default shaping function that is used in one condition only */
/* inside the BidiTransform() function.                                    */
/***************************************************************************/                                                      
int 
BidiShape(unsigned char *Buffer,
          size_t length_t,
          int Onecell,
          int OS_flag,
          int orient,
          int csd)
{
    unsigned char   state = InitialState;
    /* this variable holds the state of the state machine */
    unsigned char *current;   /* to hold characters for shaping */
    unsigned char *first_prec;
    unsigned char *sec_prec;
    unsigned char *third_prec;
    unsigned char *next; 
    unsigned char blank  = 0x20;
    unsigned char symbol = 0x21; /* ! */
    int           counter;    /* loop counter */
    int           i;
    /********************** start of processing ***************************/
    switch (csd) {
    case TEXT_SHAPED:  /* auto shaping */
        if (orient == ORIENTATION_LTR) {  /* LTR */
            /* loop on whole string to be shaped */
            for (counter = length_t - 1; counter >= 0; counter--) {
                current = &Buffer[counter];
                /* check if we have a first preceding charcater */
                if (counter < (length_t - 1))
                    first_prec = &(Buffer[counter + 1]);
                else
                    first_prec = &blank;
                /* check if we have a second preceding charcater */
                if (counter < (length_t - 2))
                    sec_prec = &(Buffer[counter + 2]);
                else
                    sec_prec = &blank;
                /* check if we have a third preceding charcater */
                if (counter < (length_t - 3))
                    third_prec = &(Buffer[counter + 3]);
                else
                    third_prec = &blank;
                /* now shape current and the preceding chars */
                csd_engine(current, first_prec, sec_prec, third_prec,
                           &state, OS_flag, Onecell);
            }
            /*
             * now assume we have a symbol after the string and reshape last
             * three chars. We do not assume a blank, because in case last
             * char is a seen, we do not want it to overwrite the blank with
             * a tail.  
             */
            third_prec = sec_prec;
            sec_prec = first_prec;
            first_prec = &(Buffer[0]);  /* this is the last char in the string */
            current = &symbol;          /* this is a hypothetical symbol */
            csd_engine(current, first_prec, sec_prec, third_prec,
                       &state, OS_flag, Onecell);
        } else {                /* RTL */
            /* loop on whole string to be shaped */
            for (counter = 0; counter < length_t; counter++) {
                current = &Buffer[counter];
                /* check if we have a firct preceding charcater */
                if (counter > 0)
                    first_prec = &(Buffer[counter - 1]);
                else
                    first_prec = &blank;
                /* check if we have a second preceding charcater */
                if (counter > 1)
                    sec_prec = &(Buffer[counter - 2]);
                else
                    sec_prec = &blank;
                /* check if we have a third preceding charcater */
                if (counter > 2)
                    third_prec = &(Buffer[counter - 3]);
                else
                    third_prec = &blank;
                /* now shape current and the preceding chars */
                csd_engine(current, first_prec, sec_prec, third_prec,
                           &state, OS_flag, Onecell);
            }
            /*
             * now assume we have a symbol after the string and reshape last
             * three chars. We do not assume a blank, because in case last
             * char is a seen, we do not want it to overwrite the blank with
             * a tail.  
             */
            third_prec = sec_prec;
            sec_prec = first_prec;
            first_prec = &(Buffer[length_t - 1]);
            current = &symbol;  /* this is a hypothetical symbol */
            csd_engine(current, first_prec, sec_prec, third_prec,
                       &state, OS_flag, Onecell);
        }
        break;
    default:                                    /* all special shapes */
        if (orient == ORIENTATION_LTR) {        /* LTR                */
            next = &(Buffer[length_t - 1]);     /* initial value      */
            /* loop on whole string to be shaped */
            for (counter = length_t - 1; counter >= 0; counter--) {
                current = next;
                /* check if we have a next charcater */
                if (counter > 0)
                    next = &(Buffer[counter - 1]);
                else
                    next = &symbol;
                csd_special(csd, current, next);
            }
        } else {                /* RTL */
            next = &(Buffer[0]);/* initial value */
            /* loop on whole string to be shaped */
            for (counter = 0; counter < length_t; counter++) {
                current = next;
                /* check if we have a next charcater */
                if (counter < (length_t - 1))
                    next = &(Buffer[counter + 1]);
                else
                    next = &symbol;
                csd_special(csd, current, next);
            }
        }
        break;
    }                           /* end switch */
    return (0);
}

/******************************************************************************/
/* FUNCTION NAME: adjust_sec_succ()                                           */
/*                                                                            */  
/* Given the original character that is correctly shaped according to its     */
/* successor, and the new character that is correctly shaped according to its */
/* predecessor, it decides the correct shape of the character.  This function */
/* is needed to determine the shape of the second succeeding character in     */
/* edit shaping . There are four possibilities :                              */
/*                1) the character was not changed at all.                    */
/*                2) the character was a space and is replaced by a vowel,    */
/*                   this is in case we had seen+vowel+space, and changed to  */
/*                   seen+tail+vowel                                          */
/*                3) the character was an alef and is changed to a vowel, this*/
/*                   is in case we had lam+vowel+alef, and is changed to      */
/*                   lam+alef+vowel Or it is an alef and chaged to a space,   */
/*                   in case we had a lam and an alef, and is changed to      */  
/*                   lam-alef + space, in Hosr mode. 4) the characetr is the  */
/*                   same char, but different shape, this is the case if we   */
/*                   had dal+vowel+middle ein, and is changed to dal+vowel+   */
/*                   initial ein.                                             */ 
/******************************************************************************/
void  adjust_sec_succ (unsigned char old_char,
                       unsigned char *new_char)
{
    if (old_char == *new_char)
        return;                                 /* case 1 */
    if ((Group(old_char) == Spaces_Group) &&    /* case 2 */
        (Group(*new_char) == Vowels_Group))
        return;
    if (((Group(old_char) == Alef_Group) ||
         (Group(old_char) == Spaces_Group)) &&  /* case 3 */
        (Group(*new_char) == Vowels_Group))
        return;
    /* Now we are sure that we are in case 4 */
    /* If new char is initial, check if it needs to be initial or isolated */
    /*
     * If it were initial or middle it means that it was left connected, so
     * since it is already delayed shaped, do nothing. If it was final or
     * isolated, it means that it is not left connected, so change it
     * accordingly. 
     */
    if (*new_char == CHRGRP[*new_char - 128][1])        /* initial         */
        if ((old_char == CHRGRP[old_char - 128][3])     /* final           */
            || (old_char == CHRGRP[old_char - 128][0])) /* or isolated     */
            *new_char = CHRGRP[old_char - 128][0];      /* set to isolated */
    if (*new_char == CHRGRP[*new_char - 128][2])        /* middle          */
        if ((old_char == CHRGRP[old_char - 128][3])     /* final           */
            || (old_char == CHRGRP[old_char - 128][0])) /* or isolated     */
            *new_char = CHRGRP[old_char - 128][3];      /* set to final    */
}

/***************************************************************************/
/* FUNCTION NAME: FindState()                                              */
/*                                                                         */
/* Given the preceeding characters, this function returns the state of the */
/* current character.                                                      */  
/***************************************************************************/
void  FindState (unsigned char *State,
                 unsigned char first_prec,
                 unsigned char sec_prec,
                 unsigned char third_prec)
{
    unsigned char            c1, c2, c3;
    unsigned char            symbol = 0x21;     /* ! */

    /* if we have a vowel we must look at the character before it */
    if (Group(first_prec) == Vowels_Group)
    {
        c1 = sec_prec;
        c2 = third_prec;
        c3 = symbol;
    } else {
        c1 = first_prec;
        c2 = sec_prec;
        c3 = third_prec;
    }
    *State = InitialState;
    switch (Group(c1))
    {
        case Symbols_Group:
        case Spaces_Group:          /* no left connection */
        case Alef_Group:
        case R_Conn_Group:
            *State = InitialState;
            break;
        case Vowels_Group:          /* just in case we have two vowels */
        case Normal_Group:
            *State = MiddleState;
            break;
        case Seen_Group:
            *State = SeenState;
            break;
        case Yeh_Hamza_Group:
            *State = YehHamzaState;
            break;
        case Lam_Group:
            if (Group(c2) <= R_Conn_Group)
                *State = LamIsoState;
            else
                *State = LamConnState;
            break;
    }
}
