/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group


******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "uc_bdstruct.h"
#include "uc_tables.h"
#include "uc_data.h"

BooleanValue    IsVN        (wchar_t xChar); 
BooleanValue    IsEnglish   (short xChar);
int             specialChar (wchar_t ch);
int             ThaiTransform();
int             VNTransform();

/***********************************************************************/
#define GRP_SIZE                   256
#define Table0                    0x10
#define Table1                    0x20
#define Table2                    0x30
#define Table3                    0x40
#define Table4                    0x50
#define Table5                    0x60
#define Table6                    0x70
#define Table7                    0x80
#define Table8                    0x90

/***************************************************************************/
/* FUNTCION NAME: isLamAlefChar()                                          */
/*                                                                         */
/* This function returns 1 for LamAlef characters 0 otherwise.             */
/***************************************************************************/
isLamAlefChar(UNI_CHAR ch) 
{
    if( (ch>=0xFEF5)&&(ch<=0xFEFC) )
        return (1);
    else
        return (0);
}
/***************************************************************************/
/* FUNTCION NAME: HandleLamAlefSpecial()                                   */
/*                                                                         */
/* This function is used in the special cases of the LamAlef.              */
/* These special cases arise from the fact that LamAlef in visual shaped   */
/* mode is only one character that occupies only one cell, while  LamAlef  */
/* in implicit nominal mode is two characters which are the Lam & Alef that*/
/* occupies two cells and not only one, so a solution has to made to enable*/
/* the expansion LamAlefs.                                                 */
/***************************************************************************/                        
int HandleLamAlefSpecial(PICSPARAMRec IcsRec)
{
    int        i          = 0;
    int        j          = 0;
    int        inSize     = IcsRec->size;
    int        outSize    = IcsRec->size;
    PUNI_CHAR  tmpBuffer  = NULL;
    UNI_CHAR   LAM        = 0x0644;

                               /* 0xFEF5  0xFEF6  0xFEF7  0xFEF8  0xFEF9  0xFEFA  0xFEFB  0xFEFC */ 
    wchar_t convertLamAlef[] = {  0x0622, 0x0622, 0x0623, 0x0623, 0x0625, 0x0625, 0x0627, 0x0627 };

    /* Calculate the needed size for the expansion & compare it with the output size */
    for(i=0;i<inSize;i++) {
        if( isLamAlefChar(IcsRec->buffer_out[i]) ) {
             outSize++;
        }
    }	

    /* 
     * To Be Done:
     * 1- We'll need to compare the user's OutSize with the calculated size.
     * 2- The process of expanding the LamAlef will need to recalculate the mappings.
     * 3- In order for this function to work properly, it'll need further modifications
     *    to be made to the usage of the *OutSize & IcsRec->size_out in the whole engine.
     */

    if(outSize > IcsRec->size_out) {
        errno            = ENOMEM;
        IcsRec->size_out = outSize;
        return(0);
    }

    tmpBuffer = (PUNI_CHAR)calloc((outSize+1),sizeof(UNI_CHAR));
    /* Fails to allocat needed memory for tmpBuffer */
    if(!tmpBuffer) {
	errno = ENOMEM;
	return(0);
    }

    i = j = 0;
    /* Loop on the whole buffer & expand the LamAlef characters */
    while(i < outSize && j < outSize) {
         if( isLamAlefChar(IcsRec->buffer_out[i]) ) {
             tmpBuffer[j] = LAM;
             tmpBuffer[j+1] = convertLamAlef[ IcsRec->buffer_out[i] - 0xFEF5 ];
             j++;
         }
         else
             tmpBuffer[j] = IcsRec->buffer_out[i];
         i++;
         j++;
    }

    /* Successfully expanded LamAlef characters into 2 characters Lam + Alef */
    /* Copy tmpBuffer to the output buffer IcsRec->buffer_out                */
    memcpy(IcsRec->buffer_out, tmpBuffer, (outSize+1)*sizeof(UNI_CHAR));
    IcsRec->size_out = outSize;

    free(tmpBuffer);
    return(1); /* Successfully Performed */
} 


/***************************************************************************/
/* FUNTCION NAME: getchtype()                                              */
/*                                                                         */
/* This function is used to find out the type of each character in the     */
/* buffer (LTR, RTL, AL, AN, EN, ... etc)                                  */
/***************************************************************************/
UBA_TYPE 
getchtype(UNI_CHAR ch_type)
{
    extern const unsigned char MasterTable[];
    extern const unsigned char ExtendedTable[][GRP_SIZE];

    unsigned char   MT_Entry, ET_Index;

    MT_Entry = MasterTable[(ch_type & 0xFF00) >> 8];
    ET_Index = MT_Entry >> 4;
    if (ET_Index == 0)
        return (UBA_TYPE) MT_Entry;
    else
        return (UBA_TYPE) (ExtendedTable[ET_Index - 1][ch_type & 0xff]);
}

/****************************************************************************/
/* FUNTCION NAME: ucics()                                                   */
/*                                                                          */
/* This function performs the implicit process which determines the implicit*/
/* level of the current character in the input text.                        */
/****************************************************************************/
void 
ucics(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{

    int             i = 0;
    ushort          newIS, Special, newIL, sCond;


    newIS = impTab[ucb->impSta][ucb->xType];
    Special = newIS >> 5;
    newIS = newIS & 0x1F;
    newIL = impTab[newIS][ITIL];

    if (Special > 0)
        switch (Special) {
        case 1:         /* note a: set characters to level 0 from condPos until last */
            for (i = ucb->condPos; i < ucb->ix; i++) {
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0xC000;
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] | (ucb->curLev + ucb->addLev);
            }
            ucb->condPos = -1;
            break;

        case 2:         /* note b: set characters to level 1 from condPos until last */
            for (i = ucb->condPos; i < ucb->ix; i++) {
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0xC000;
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] | (ucb->curLev + 1 + ucb->addLev);
            }
            ucb->condPos = -1;
            break;

        case 3:         /* note c: set characters to level 1 from 
                         * condPos until next to last and set last
                         * character to level 2 */
            for (i = ucb->condPos; i < ucb->ix; i++) {
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0xC000;
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] | (ucb->curLev + 1 + ucb->addLev);
            }
            IcsRec->TrgToSrcMap[i] += ucb->curLev + ucb->addLev + 2;
            ucb->condPos = -1;
            break;

        case 4:         /* note d: set condPos at the current
                         * character position  */
            ucb->condPos = ucb->ix;
            break;

        case 5:         /* note e: mark that there is no conditional
                         * string  */
            ucb->condPos = -1;
            break;

        }

    sCond = impTab[newIS][ITCOND];
    if (sCond == 0) {
        if (ucb->condPos > -1) {
            for (i = ucb->condPos; i < ucb->ix; i++) {
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0xC000;
                IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] |
                    (ucb->curLev + ucb->addLev + newIL);
                if (((ucb->curLev + ucb->addLev + newIL) % 2 == 0) &&
                    ((IcsRec->buffer_in[i] != 0x206C) ||
                     (IcsRec->buffer_in[i] != 0x206F)))
                    /* clear bit 15 */
                    IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0x7FFF;
            }
        }
        ucb->condPos = -1;
        if (ucb->linsepPos >= 0) {
            IcsRec->TrgToSrcMap[ucb->linsepPos] = 0;
            ucb->linsepPos = -1;
        }
    } else if (ucb->condPos == -1)
        ucb->condPos = ucb->ix;

    ucb->impLev = newIL;
    ucb->impSta = newIS;
    ucb->wTarget = ucb->wTarget | (ucb->curLev + ucb->addLev + ucb->impLev);
}

/************************************************************************/
/* FUNCTION NAME: BaseLvl()                                             */
/*                                                                      */
/* This function determines the base level of the input text.           */
/*  (whether LTR or RTL.)                                               */
/************************************************************************/
void 
BaseLvl(PICSPARAMRec IcsRec,
        PUCB2V ucb)
{

    UBA_TYPE        chtype;

    ucb->basLev     = 0;

    if (IcsRec->orient_in == ORIENTATION_RTL)   /* 0 = LTR, 1 = RTL */
        ucb->basLev = RTL;
    else if ((IcsRec->orient_in != ORIENTATION_LTR) &&
             (IcsRec->orient_in != ORIENTATION_RTL)) {
        while (ucb->ix < IcsRec->size) {
            chtype = getchtype(IcsRec->buffer_in[ucb->ix]);
            if (chtype == UBAT_R) {
                ucb->basLev = 1;
                break;
            }
            if (chtype == UBAT_L)
                break;
            ucb->ix++;
        }
    }
    ucb->curLev = ucb->basLev;
    ucb->impLev = 0;
    if (ucb->basLev == 0)
        ucb->impSta = 0;
    else
        ucb->impSta = 9;

    ucb->condPos = -1;
    ucb->linsepPos = -1;

    if ((ucb->basLev == 0) && (ucb->outLev == RTL))
        ucb->addLev = 2;
    else
        ucb->addLev = 0;

}

/**********************************************************************************/
/* FUNCTION NAME: pass1()                                                         */
/*                                                                                */
/* This function does the calculation of the implicit level for each character    */
/* in the input buffer buffer_in. The function also flags the positions of Arabic */
/* digits that needs transformation to Indic digits, non-spacing characters,      */
/* and characters that needs symmetric swapping.                                  */
/**********************************************************************************/
void 
pass1(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{

    UBA_TYPE        cType;
    int             savIL;

    if (ucb->ix == 0)
        ucb->pType = 0;

    ucb->wTarget = 0;

    cType = getchtype(IcsRec->buffer_in[ucb->ix]);
    switch (cType) {
        /* Bidi Special Codes */
    case UBAT_BD:

        switch (IcsRec->buffer_in[ucb->ix]) {
        case LRE:
            /* do embedding process using arg 0,0 */
            embeddingProcess(IcsRec, ucb, 0, 0);
            break;
        case RLE:
            /* do embedding process using arg 1,0 */
            embeddingProcess(IcsRec, ucb, 1, 0);
            break;
        case LRO:
            /* do embedding process using arg 0,1 */
            embeddingProcess(IcsRec, ucb, 0, 1);
            break;
        case RLO:
            /* do embedding process using arg 1,1 */
            embeddingProcess(IcsRec, ucb, 1, 1);
            break;
        case PDF:
            /* do embedding process for UBAT_S  */
            ucb->xType = UBAT_S;
            ucics(IcsRec, ucb);

            if (ucb->noOp == 0) {
                if (ucb->curLev == ucb->basLev)
                     /* Error code */ ;
                else {
                    ucb->impSta = ucb->Lstack[ucb->curLev].savIS;
                    ucb->over = ucb->Lstack[ucb->curLev].savOV;
                    ucb->curLev = ucb->Lstack[ucb->curLev].oldCL;
                    ucb->impLev = impTab[ucb->impSta][ITIL];    /* According to the implicit State
                                                                 * in the implicit table */
                    ucb->wTarget = ucb->curLev + ucb->impLev + ucb->addLev;
                }
            } else {
                ucb->noOp--;
                if (ucb->noOp == 0)
                    ucb->impSta = ucb->Lstack[ucb->curLev].savIS;
            }
            ucb->araLet = 0;
            break;

        case 0x206C:            /* Arabic Shaping Inhibit */
            if (IcsRec->boc_flag == 0xFFFF)
                ucb->Compac = 1;
            if (IcsRec->boc_flag != UNI_NULL)
                ucb->wTarget = ucb->wTarget | 0x8000;
            IcsRec->formShp = 0;
            break;

        case 0x206D:            /* Arbabic Shaping Activate */
            if (IcsRec->boc_flag == 0xFFFF)
                ucb->Compac = 1;
            if (IcsRec->boc_flag != UNI_NULL)
                ucb->wTarget = ucb->wTarget | 0x8000;
            IcsRec->formShp = 1;
            break;

        case 0x206E:            /* Numeric Shape National */
            if (IcsRec->boc_flag == 0xFFFF)
                ucb->Compac = 1;
            if (IcsRec->boc_flag != UNI_NULL)
                ucb->wTarget = ucb->wTarget | 0x8000;
            IcsRec->num_flag = NUMERALS_NATIONAL;
            break;

        case 0x206F:            /* Numeric Shape Nominal */
            if (IcsRec->boc_flag == 0xFFFF)
                ucb->Compac = 1;
            if (IcsRec->boc_flag != UNI_NULL)
                ucb->wTarget = ucb->wTarget | 0x8000;
            IcsRec->num_flag = NUMERALS_NOMINAL;
            break;


        }
        break;

        /* Block Separator */
    case UBAT_B:
        /* do implicit process for UBAT_B */
        ucb->xType = UBAT_B;
        ucics(IcsRec, ucb);

        ucb->wTarget = 0;
        /* redo Determination of the base level basLev */
        BaseLvl(IcsRec, ucb);
        break;

        /* Segment Separator */
    case UBAT_S:
        /* do implicit process for UBAT_S */

        ucb->wTarget = ucb->basLev;

        /* HMS */
        ucb->xType = UBAT_S;
        ucics(IcsRec, ucb);
        break;

    default:
        /* source character is Line Separator */
        if (IcsRec->buffer_in[ucb->ix] == 0x2028) {
            ucb->linsepPos = ucb->ix;
            ucb->wTarget = 0;
            break;
        }
        /* source character is LRM or RLM */
        if ((IcsRec->buffer_in[ucb->ix] == 0x200E) ||
            (IcsRec->buffer_in[ucb->ix] == 0x200F)) {
            if (IcsRec->boc_flag == 0xFFFF)
                ucb->Compac = 1;
            else if (IcsRec->boc_flag != UNI_NULL)
                ucb->wTarget = ucb->wTarget | 0x8000;
        }
        if (ucb->over == 1) {
            if (ucb->curLev % 2 == 1)
                cType = UBAT_R;
            else
                cType = UBAT_L;
        }
        if (((cType == UBAT_EN))
            && (IcsRec->num_flag == NUMERALS_CONTEXTUAL)) {
            if (ucb->pType == UBAT_AN) {
                ucb->wTarget = ucb->wTarget | 0xa000;
                cType = UBAT_AN;
            } else if ((ucb->pType == UBAT_W || 
                        ucb->pType == UBAT_N || ucb->pType == UBAT_CS || 
                        ucb->pType == UBAT_ES ) && 
                       (ucb->ix - 1 < IcsRec->size) && ((IcsRec->TrgToSrcMap[ucb->ix - 1] & 0x2000) == 0)) {
                int             i1 = 0;
                for (i1 = ucb->ix - 2; i1 >= 0; i1--) {
                    ucb->pType = getchtype(IcsRec->buffer_in[i1]);
                    if (((ucb->pType == UBAT_R) || ((IcsRec->TrgToSrcMap[i1] & 0x2000) != 0))){
                      /* (  && ucb->araLet == 1)  */
                        ucb->wTarget = ucb->wTarget | 0xa000;
                        cType = UBAT_AN;
                        break;
                    } else if ((ucb->pType == UBAT_L) || ((ucb->pType == UBAT_EN) &&
                               ((IcsRec->TrgToSrcMap[i1] & 0x2000) == 0))) {
                        break;
                    } else {
                        continue;
                    }

                }
            } else 
	    if (ucb->ix > 0 && (((IcsRec->TrgToSrcMap[ucb->ix-1] & 0x2000) != 0) || (ucb->pType == UBAT_R)) && ucb->araLet == 1) {
                ucb->wTarget |= 0xa000;
                cType = UBAT_AN;
            } else { /* Added to solve the contextual number problem in RTL */
              if(ucb->ix == 0 && IcsRec->orient_in != IcsRec->orient_out){
                ucb->wTarget |= 0xa000;
                cType = UBAT_AN;
              } else if (ucb->ix > 0) {
                int  i1 = 0, flag=0;
                for (i1 = ucb->ix - 2; i1 >= 0; i1--) {
                  ucb->pType = getchtype(IcsRec->buffer_in[i1]);
                  if (((ucb->pType == UBAT_R) || ((IcsRec->TrgToSrcMap[i1] & 0x2000) != 0)) && ucb->araLet == 1) {
                    ucb->wTarget = ucb->wTarget | 0xa000;
                    cType = UBAT_AN;
                    flag = 1;
                    break;
                  } else if ((ucb->pType == UBAT_L) || ((ucb->pType == UBAT_EN) &&
                             ((IcsRec->TrgToSrcMap[i1] & 0x2000) == 0))) {
                    break;
                  } else {
                    continue;
                  }
                  
                }
              }
           }
        }
        if (cType == UBAT_L)
            ucb->araLet == 0;

        if (cType == UBAT_R) {
            if ((IcsRec->buffer_in[ucb->ix] >= 0x0600) &&
                (IcsRec->buffer_in[ucb->ix] <= 0x06EF)) {

                ucb->araLet = 1;
                if (IcsRec->txtShp_flag != FALSE && IcsRec->txtShp_flag != TEXT_NOMINAL)
                    ucb->Shaping = 1;
            }
            if ((IcsRec->buffer_in[ucb->ix] >= 0xFB50) &&
                (IcsRec->buffer_in[ucb->ix] <= 0xFEFC)) {

                ucb->araLet = 1;
                if ((IcsRec->txtShp_flag != FALSE && IcsRec->txtShp_flag != TEXT_NOMINAL) &&
                    (IcsRec->formShp == 1)) {
                    ucb->Shaping = 1;
                    ucb->wTarget = ucb->wTarget | 0x8000;
                }
            }
        }
      
        if (cType == UBAT_ET) {
            int   iFlag = 0, i2 = 0;
            if (ucb->pType == UBAT_EN)
                cType = UBAT_EN;
            else if (ucb->pType == UBAT_AN)
                cType = UBAT_N;
            while (iFlag == 0) {
                if (ucb->ix + i2 + 1 < IcsRec->size)
                    ucb->pType = getchtype(IcsRec->buffer_in[ucb->ix + i2 + 1]);
            else {
                  iFlag = 1;
                  continue;
            }
            if (ucb->pType == UBAT_EN) {
               if (ucb->araLet == 1)
                   cType = UBAT_N;
               else
                   cType = UBAT_EN;
               iFlag = 1;
            }
            else if (ucb->pType == UBAT_AN) {
                    cType = UBAT_N;
                    iFlag = 1;
            }
            if (ucb->pType == UBAT_ET) {
                i2++;
                continue;
            } else
                iFlag = 1;
            }
        }
        if (cType == UBAT_ES || cType == UBAT_CS) {
            ucb->xType =(ucb->ix +1 < IcsRec->size)? getchtype(IcsRec->buffer_in[ucb->ix+1]):0;
            if (ucb->pType == UBAT_EN) {
                ucb->pType = (ucb->ix + 1 < IcsRec->size) ? getchtype(IcsRec->buffer_in[ucb->ix + 1]) : 0;
                cType = (ucb->pType == UBAT_EN) ? ucb->pType : UBAT_N;
            } else if (ucb->pType == UBAT_AN && cType == UBAT_CS) {
                ucb->pType = (ucb->ix + 1 < IcsRec->size) ? getchtype(IcsRec->buffer_in[ucb->ix + 1]) : 0;
                cType = (ucb->pType == UBAT_EN) ? UBAT_AN : UBAT_N;
            } 
                 
            else
                cType = UBAT_N;
        }
        savIL = ucb->impLev;
        /* do implicit Process for cType */
        ucb->xType = cType;
        ucics(IcsRec, ucb);
        if ((cType == UBAT_EN)
            && ((IcsRec->num_flag == NUMERALS_NATIONAL)
                )) {
            ucb->wTarget = ucb->wTarget | 0xa000;
            cType = UBAT_AN;
        }

        /* Find if character is spacing */
        if (UCQSPAC(cType) == FALSE)
            ucb->wTarget = (ucb->curLev + ucb->addLev + savIL) | 0x4000;

        if ((cType == UBAT_N) &&
            (IcsRec->symmetric == TRUE) &&
            ((ucb->condPos > -1) ||
             ((ucb->curLev + ucb->impLev) % 2 != 0)))
            ucb->wTarget = ucb->wTarget | 0x8000;

        break;                  /* Case End */
    }

    ucb->pType = cType;
    IcsRec->TrgToSrcMap[ucb->ix] = ucb->wTarget;        /* put wTarget in target area at position ix */
}

/***************************************************************************/
/* FUNTCION NAME: embeddingProcess()                                       */
/*                                                                         */
/***************************************************************************/
int 
embeddingProcess(PICSPARAMRec IcsRec, PUCB2V ucb, int levelEvenOdd, int newOver)
{
    if (levelEvenOdd == 0) {
        ucb->newCL = ((IcsRec->TrgToSrcMap[ucb->ix-1]+1)%2 == 0 ? (IcsRec->TrgToSrcMap[ucb->ix-1]+1):(IcsRec->TrgToSrcMap[ucb->ix-1]+2));
        ucb->xType = UBAT_L;
    } else {
        ucb->newCL = (IcsRec->TrgToSrcMap[ucb->ix - 1] + 1) | 1;
        ucb->xType = UBAT_R;
    }

    ucics(IcsRec, ucb);

    if (ucb->noOp == 0)
        ucb->Lstack[ucb->newCL].savIS = ucb->impSta;
    ucb->Lstack[ucb->newCL].savOV = ucb->over;
    if (ucb->newCL < 15) {
        ucb->Lstack[ucb->newCL].oldCL = ucb->curLev;
        ucb->curLev = ucb->newCL;
        ucb->impLev = 0;
        ucb->impSta = levelEvenOdd * 9;
        ucb->over = newOver;
        ucb->wTarget = ucb->curLev + ucb->impLev + ucb->addLev;
    } else {
        ucb->noOp++;
    }
    ucb->araLet = 0;
}

/*********************************************************************/
/* FUNCTION NAME: pass2()                                            */
/*                                                                   */
/* pass2() uses the level of each char to invert segments of         */
/* contiguous characters which have the same level. This is done by  */
/* reversing the segments of highest level and decrementing the level*/
/* of each char in those segments, and iterating this operation until*/
/* all chars have the same odd level.                                */
/*********************************************************************/
void 
pass2(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{

    int             lolim[19];
    int             hilim[19];
    int             splolim             = 0;
    int             sphilim             = -1;
    int             i, x, xLev, lastLev = 0;
    int             iy, wLev;

    for (x = 0; x < 19; x++) {
        lolim[x] = 0;
        hilim[x] = IcsRec->size - 1;
    }

    for (ucb->ix = 0; ucb->ix < IcsRec->size; ucb->ix++) {
        /* retain only the level */
        xLev = IcsRec->TrgToSrcMap[ucb->ix] & 0x1FFF;
        /* set limits of inverse for not-spacing if needed */
        if ((xLev & 1) != ucb->outLev) {
            if (ucb->ix <= sphilim)     /* ix belongs to defined run */
                /* reverse not-spacing */
                x = splolim + sphilim - ucb->ix;
            else {
                if ((ucb->ix + 1 < IcsRec->size) && ((IcsRec->TrgToSrcMap[ucb->ix] & 0x4000) == 0) &&
                    ((IcsRec->TrgToSrcMap[ucb->ix + 1] & 0x4000) != 0)) {
                    int             counter = 0;

                    splolim = ucb->ix;
                    for (counter = splolim + 1; counter < IcsRec->size; counter++) {
                        if ((IcsRec->TrgToSrcMap[counter] & 0x4000) != 0)
                            continue;
                        else
                            break;
                    }
                    sphilim = counter - 1;
                    x = splolim + sphilim - ucb->ix;
                } else
                    x = ucb->ix;
            }
        } else
            x = ucb->ix;
        /*
         * check the limits : if the last character had a level (lastLev)
         * greater than or equal to xLev, its segment is contained in the
         * segment of ix, and all boundaries are already known; if xLev is
         * greater than lastLev, this is the start of a new segment, lolim
         * and hilim must be updated  
         */
        if (xLev > lastLev) {
            wLev = xLev;
            iy = ucb->ix + 1;
            while (wLev > lastLev) {
                lolim[wLev] = ucb->ix;
                while (iy < IcsRec->size) {
                    if ((IcsRec->TrgToSrcMap[iy] & 0x3FFF) < wLev)
                        break;
                    iy++;
                }

                hilim[wLev] = iy - 1;
                wLev--;
            }
        }
        /* Reverse according to levels */
        for (i = xLev; i >= (ucb->outLev + 1); i--)
            x = lolim[i] + hilim[i] - x;

        /* Update target area */
        IcsRec->TrgToSrcMap[ucb->ix] = IcsRec->TrgToSrcMap[ucb->ix] & 0xa000;
        IcsRec->TrgToSrcMap[ucb->ix] = IcsRec->TrgToSrcMap[ucb->ix] + x;
        lastLev = xLev;
    }
}
/**************************************************************************/
/* FUNCTION NAME: pass3()                                                 */
/*                                                                        */
/* Replaces the content of the target area with positions such that each  */
/* character in the target area will receive the position of the source   */
/* character that will map to it.                                         */
/**************************************************************************/
void 
pass3(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{
    ushort          logPos, logPtr, visPos, wPtr;

    for (ucb->ix = 0; ucb->ix < IcsRec->size; ucb->ix++) {
        logPos = ucb->ix;
        logPtr = IcsRec->TrgToSrcMap[ucb->ix];
        while ((logPtr & 0x4000) == 0) {
            visPos = logPtr & 0x1FFF;
            wPtr = IcsRec->TrgToSrcMap[visPos];
            if ((wPtr & 0x4000) > 0)
                break;
            IcsRec->TrgToSrcMap[visPos] = logPos + (logPtr & 0xa000) + 0x4000;
            logPos = visPos;
            logPtr = wPtr;
        }
    }
}
/**************************************************************************/
/* FUNCTION NAME: pass4()                                                 */
/*                                                                        */
/* Replaces the content of the target area with the source characters from*/
/* the indicated position. European digits which need Arabic-indic shapes */
/* translated; Arabic presentation forms which need shaping are replaced  */
/* by the corresponding nominal letter.                                   /*
/**************************************************************************/
void 
pass4(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{
    int             i;
    int             tmp;
    int             j             = 0;
    ushort          xTran;
    UNI_CHAR        xchar;

    const unsigned short TashkeelChar[] = { 0x64B, 0x64B, 0x64C, 0x1A , 0x64D, 0x1A , 0x64E, 0x64E,
                                            0x64F, 0x64F, 0x650, 0x650, 0x651, 0x651, 0x652, 0x652 };

    for (i = 0; i < IcsRec->size; i++) {
        xTran = IcsRec->buffer_out[i] & 0x8000;
        if (IcsRec->order_flag == TRUE)
            xchar = IcsRec->buffer_in[IcsRec->buffer_out[i] & 0x1FFF];
        else
            xchar = IcsRec->buffer_in[i];
        if (xTran > 0) {
            if ((xchar==0x200E) || (xchar==0x200F) || ((xchar>=0x202A) && (xchar<=0x202E)) || ((xchar>=0x206C) && (xchar<=0x206F)))
                xchar = IcsRec->boc_flag;
            else if ((xchar >= 0x0030) && (xchar <= 0x0039) && (IcsRec->buffer_out[i] & 0x8000) != 0)
                xchar = xchar + (0x0660 - 0x0030);
            else {
                if (IcsRec->symmetric != 0)
                    xchar = UCQSYMM(xchar);
                if ((IcsRec->orient_in != IcsRec->orient_out)) {
                    int             i1=0, j=0;
                    UBA_TYPE        cType;
                    if (j < IcsRec->size)
                        i1 = IcsRec->buffer_out[i] & 0x1FFF;
                    j = i;
                    do {
                        cType = getchtype(IcsRec->buffer_in[i1]);
                        if (cType == UBAT_R)
                            break;
                        else if ((cType == UBAT_L || cType == UBAT_EN) 
                                 && ucb->basLev % 2 == 0) {
                            xchar = UCQSYMM(xchar);
                            break;
                        } else {
                            j++;
                            if (j < IcsRec->size)
                                i1 = IcsRec->buffer_out[j] & 0x1FFF;
                        }
                    } while (j <= IcsRec->size);
                }
            }
        }
        if (IcsRec->Tables == TEXT_IMPLICIT) {
            if((xchar >= 0xFE70) && (xchar <= 0xFE7F)) {
                tmp = xchar - 0xFE70;
                xchar = TashkeelChar[tmp];
            }
            if((xchar >= 0xFE80) && (xchar <= 0xFEF4)) {
                tmp = xchar - 0xFE80;
                xchar = ComtoNom[tmp];
            }
        }
        if((IcsRec->num_flag == NUMERALS_NOMINAL) && (xchar >= 0x0660) && (xchar <= 0x0669))
                xchar = xchar - (0x0660 - 0x0030);

        if((xchar == 0x0625) && (strcmp(IcsRec->codeset,"IBM420") == 0) )
                xchar = 0x0627;
        IcsRec->buffer_out[i] = xchar;
    }
    /* Call HandleLamAlefSpecial() to expand LamAlef into Lam + Alef */
    HandleLamAlefSpecial(IcsRec);
}

/**************************************************************************/
/* FUNCTION NAME: InvertMap()                                             */
/*                                                                        */
/* This function is used to invert the buffer between lowlimit & highlimit*/
/**************************************************************************/
void InvertMap(PICSPARAMRec IcsRec,int lowlimit,int highlimit)
{
        int             i = 0;
        int             j = 0;
        UNI_CHAR        tempchar;

        for(i=lowlimit,j=highlimit;i<j;i++,j--) {
            tempchar = IcsRec->buffer_out[i];
            IcsRec->buffer_out[i] = IcsRec->buffer_out[j];
            IcsRec->buffer_out[j] = tempchar;
        }
}

/**************************************************************************/
/* FUNCTION NAME: pass5()                                                 */
/*                                                                        */
/* This is the main shaping function.                                     */
/**************************************************************************/
void 
pass5(PICSPARAMRec IcsRec,
      PUCB2V ucb)
{
    /*
     * Link attribute are unsigned 16 bit integer with the following format:
     * bit 0 :     links to right           LINKR      = 1 
     * bit 1 :     links to leift           LINKL      = 2 
     * bit 2 :     is irrelevant to linking IRRELEVANT = 4 
     * bit 4 :     Lam type                 LAMTYPE    = 16
     * bit 5 :     Alef type                ALEFTYPE   = 32 
     * bits 8-15 : offset to presentation shapes starting at 0xFE70 
     */

    int             i, iend, step;
    int             prevPos, lastPos, Nx, Nw;
    int             flag;
    unsigned int    Shape;
    unsigned short  prevLink = 0, lastLink = 0, currLink, nextLink;
    UNI_CHAR        wLamalef;
    int             size;

    const unsigned  short shapeTable[5][4][4] = { {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 1, 0, 3}, {0, 1, 0, 1}},
                                                  {{0, 0, 2, 2}, {0, 0, 1, 2}, {0, 1, 1, 2}, {0, 1, 1, 3}},
                                                  {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 1, 0, 3}, {0, 1, 0, 3}},
                                                  {{0, 0, 1, 2}, {0, 0, 1, 2}, {0, 1, 1, 2}, {0, 1, 1, 3}}};

    size = IcsRec->size;
    if( (IcsRec->orient_in == ORIENTATION_RTL) &&
        (IcsRec->orient_out == ORIENTATION_RTL))
        InvertMap(IcsRec,0,size-1);
    if (ucb->outLev == RTL && IcsRec->orient_in != IcsRec->orient_out) {
        i = 0;
        iend = size;
        step = 1;
    } else {
        i = size - 1;
        iend = -1;
        step = -1;
    }
    currLink = uba_getLink(IcsRec->buffer_out[i]);
    prevPos = i;
    lastPos = i;
    Nx      = -2, Nw = 0;
    while (i != iend) {
        if ((currLink & 0xFF00) > 0 || IcsRec->buffer_out[i] == 0x064B) { /* more than one shape */
            Nw = i + step;
            while (Nx < 0) {    /* we need to know about next char */
                if (Nw == iend) {
                    nextLink = 0;
                    Nx = 3000;
                } else {
                    nextLink = uba_getLink(IcsRec->buffer_out[Nw]);
                    if ((nextLink & IRRELEVANT) == 0)
                        Nx = Nw;
                    else
                        Nw = Nw + step;
                }
            }
            if (((currLink & ALEFTYPE) > 0) && ((lastLink & LAMTYPE) > 0) 
                && (IcsRec->compc == TRUE)) {
                wLamalef = Lamalef(IcsRec->buffer_out[i]);
                if (wLamalef != UNI_NULL) {
                    if (ucb->outLev == 1 /* RTL device */
                        && IcsRec->orient_in != IcsRec->orient_out) {
                        /* To fix the space problem */
                        int             dummy = i;
                        if (IcsRec->backward != TRUE){
                          IcsRec->buffer_out[lastPos] = wLamalef;
                          while (dummy < IcsRec->size - 1) {/* drop the Alef */
                            IcsRec->buffer_out[dummy] = 
                                    IcsRec->buffer_out[dummy + step];
                            if (IcsRec->compc == TRUE)
                              IcsRec->TrgToSrcMap[dummy] = 
                                      IcsRec->TrgToSrcMap[dummy + step];
                            dummy += step;
                          }
                          size--;
                          i--;
                          IcsRec->buffer_out[size] = 0;
                          IcsRec->buffer_out[i] = wLamalef;
                        } else {
                          IcsRec->buffer_out[lastPos] = 0x20;
                          IcsRec->buffer_out[dummy] = wLamalef;
                          ucb->Compac = 1;
                        }
                    } else { /* LTR device */
                        /* To fix the space in front of the LamAlef */
                        int             dummy = i;
                        if (IcsRec->backward != TRUE) {
                        while (dummy != size - 1) {
                            IcsRec->buffer_out[dummy] = /* drop the Lam */
                                    IcsRec->buffer_out[dummy - step];
                            IcsRec->TrgToSrcMap[dummy] = 
                                    IcsRec->TrgToSrcMap[dummy - step];
                            dummy -= step;
                        }
                        size--;
                        IcsRec->buffer_out[size] = 0;
                        IcsRec->buffer_out[i] = wLamalef;
                        } else {
                          IcsRec->buffer_out[dummy] = 0x20;
                          IcsRec->buffer_out[lastPos] = wLamalef;
                          ucb->Compac = 1; 
                        }
                        /* move irrelevant characters */
                    }
                }
                lastLink = prevLink;
                currLink = uba_getLink(wLamalef);
                if (ucb->Compac == 1){
                  Shape = shapeTable[nextLink & (LINKR + LINKL)]
                    [lastLink & (LINKR + LINKL)]
                    [currLink & (LINKR + LINKL)];
                  IcsRec->buffer_out[lastPos] = 
                          0xFE70 + (currLink >> 8) + Shape;
                }
            }
            /*
             * get the proper shape according to link ability of neighbors
             * and of character; depends on the order of the shapes
             * (isolated, initial, middle, final) in the compatibility area 
             */

            flag = specialChar(IcsRec->buffer_out[i]);

            if (IcsRec->txtShp_flag == TEXT_SHFORM1) {
                if (flag == 0)
                    Shape = 2;
                else
                    Shape = 0;
            } else if (IcsRec->txtShp_flag == TEXT_SHFORM2) {
                if (flag == 0)
                    Shape = 3;
                else
                    Shape = 1;
            } else if (IcsRec->txtShp_flag == TEXT_SHFORM3) {
                if (flag == 0)
                    Shape = 1;
                else
                    Shape = 1;
            } else if (IcsRec->txtShp_flag == TEXT_SHFORM4)
                Shape = 0;
            else {
                Shape = shapeTable[nextLink & (LINKR + LINKL)]
                                  [lastLink & (LINKR + LINKL)]
                                  [currLink & (LINKR + LINKL)];
                if (flag == 1)
                    Shape = (Shape == 1 || Shape == 3) ? 1 : 0;
                else if (flag == 2){   /* We have a tashkeel character */
                  if ((lastLink & LINKL) && (nextLink & LINKR)
                      && (IcsRec->buffer_out[i] != 0x64C)
                      && (IcsRec->buffer_out[i] != 0x64D))
                    Shape = 1;
                  else{
                    Shape = 0;
                    currLink -= (LINKL+LINKR);
                    if(lastLink & LINKL)
                      IcsRec->buffer_out[i-step] -=2;
                  }
                }
            }
            if (ucb->Compac == 0)
              IcsRec->buffer_out[i] = 0xFE70 + (currLink >> 8) + Shape;
            ucb->Compac =0;

        }
        /* move one notch forward */
        if ((currLink & IRRELEVANT) == 0) {
            prevLink = lastLink;
            lastLink = currLink;
            prevPos = lastPos;
            lastPos = i;
        }
        i = i + step;
        if (i == Nx) {
            currLink = nextLink;
            Nx = -2;
        } else if(i != iend) 
            currLink = uba_getLink(IcsRec->buffer_out[i]);
    }
    if( (IcsRec->orient_in == ORIENTATION_RTL) &&
        (IcsRec->orient_out == ORIENTATION_RTL))
        InvertMap(IcsRec,0,size-1);
    IcsRec->size_out = size;
}

/**************************************************************************/
/* FUNCTION NAME: UCB2VIS()                                               */
/*                                                                        */
/* In this function the actual transformation takes place by calling the  */
/* five layout engine passes                                              */
/**************************************************************************/
int 
UCB2VIS(PICSPARAMRec IcsRec)
{

    int             i = 0, j = 0, RC = 0;
    int             x = 0;
    int             iend, step;
    int             loop  = 0;
    UCB2V           ucb;
    LstackRec       Lstack[16];
    MultiLingual    Start = { 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
    pMultiLingual   Slider;
    unsigned short  prevLink  = 0, lastLink = 0, currLink = 0, nextLink;

    IcsRec->size_out = 0;
    /*******************/
    /* Initializations */
    /*******************/
    ucb.ix           = 0;
    ucb.Compac       = 0;
    ucb.Shaping      = 0;
    ucb.Lstack       = Lstack;
    ucb.indexOfStack = 0;
    ucb.over         = 0;
    ucb.noOp         = 0;
    ucb.araLet       = 0;

    if (IcsRec->orient_out == ORIENTATION_LTR)
        ucb.outLev = LTR;
    else
        ucb.outLev = RTL;

    Slider = &Start;

    /*----------------------------------------*/
    /* Determination of the base level basLev */
    /*----------------------------------------*/
    BaseLvl(IcsRec, &ucb);

    ucb.ix = 0;
    while (ucb.ix < IcsRec->size) {
        pass1(IcsRec, &ucb);
        ucb.ix++;
    }

    /* do Implicit process for UBAT_B to resolve possible conditional string */
    ucb.xType = UBAT_B;
    ucics(IcsRec, &ucb);
    for (i = 0; i < IcsRec->size; i++) {
        if (IcsRec->buffer_in[i] >= LRE && IcsRec->buffer_in[i] <= RLO) {
            for (j = i; j < IcsRec->size - 1; j++) {
                IcsRec->buffer_in[j] = IcsRec->buffer_in[j + 1];
                IcsRec->TrgToSrcMap[j] = IcsRec->TrgToSrcMap[j + 1];
            }
            IcsRec->size--;
            i--;
        }
    }
    if (IcsRec->A_level) {
        for (i = 0; i < IcsRec->size; i++)
            IcsRec->A_level[i] = IcsRec->TrgToSrcMap[i] & 0x00FF;
    }
    if (IcsRec->order_flag == TRUE) {
        /*
         * Pass 2: This pass must not be executed when there is no request
         *         for reordering 
         */
        pass2(IcsRec, &ucb);

        /*
         * Pass 3: The logical to visual mapping must be converted to a visual-to-logical mapping.
	 * We use bit 14 in each position of the target area to mark positions which have been processed.  
         */
        pass3(IcsRec, &ucb);

        for (i = 0; i < IcsRec->size; i++)
            IcsRec->buffer_out[i] = IcsRec->TrgToSrcMap[i];
        /*
         * Pass 4: Replaces the content of the target area set by pass3 with the source charcters from the 
	 * indicated position. European digits which need Arabic-Indic shapes are translated; Arabic presentation
         * forms which need shaping are replaced by the corresponding nominal letters; symbols which need swapping 
	 * are replaced by thier symmetric symbol; bidi special codes which need replacement are replaced. 
         */
        pass4(IcsRec, &ucb);
    }

    if (ucb.outLev == RTL) {
        i    = 0;
        iend = IcsRec->size;
        step = 1;
    } else {
        i    = IcsRec->size - 1;
        iend = -1;
        step = -1;
    }

    /* To fill in the property MSB */
    if (IcsRec->A_level) {
        IcsRec->A_level[iend - step] += 0x80;
        if(0<(iend-2*step) && IcsRec->size>(iend-2*step))
        for (j = iend - (2 * step); j != i ; j -= step) {
            if (j < 0 || j > IcsRec->size)
                break;
            if ((currLink = uba_getLink(IcsRec->buffer_in[j + step]) & LAMTYPE) && (IcsRec->compc == TRUE)) {
                if ((lastLink = uba_getLink(IcsRec->buffer_in[j]) & ALEFTYPE)) {
                    IcsRec->A_level[j] &= 0x0f;
                } else
                    IcsRec->A_level[j] += 0x80;
            } else {
                IcsRec->A_level[j] += 0x80;
            }
        }
        if( iend != i+step)
        currLink = uba_getLink(IcsRec->buffer_in[i + step]);
        lastLink = uba_getLink(IcsRec->buffer_in[i]);
        if ((currLink & LAMTYPE) && (lastLink & ALEFTYPE) && (IcsRec->compc == TRUE))
            IcsRec->A_level[i] &= 0x0F;
        else
            IcsRec->A_level[i] += 0x80;
    }
    lastLink = 0;

    /*
     * Pass 5 : Excuted only if Shaping is not zero and it replaces Arabic
     *          base letters by the appropriate presentation forms.  
     */
    if ((IcsRec->txtShp_flag != 0) && (IcsRec->txtShp_flag != TEXT_NOMINAL))
        pass5(IcsRec, &ucb);

    x = getLang(Slider, 0, IcsRec);
    if(x == -1) {
            if(errno = ENOMEM)
                return(-1);
    }
    if (x == 1) {
        int             x2 = 0;
        Slider = &Start;
        loop = 0;
        do {
           if (!(Slider->lang)) {
               if (!(Slider->next))
                   break;
               else {
                   Slider = Slider->next;
                   loop = 1;
                   continue;
               }
           }
           x2 = Slider->end - Slider->start + 1;

           Slider->dummy = (wchar_t *) calloc(((x2 * 3) + 1), sizeof(wchar_t));
           Slider->i_to_o = (size_t *) calloc((x2), sizeof(size_t));
           Slider->o_to_i = (size_t *) calloc((x2 * 3), sizeof(size_t));
           Slider->prop = (unsigned char *) calloc((x2), sizeof(char));
           if (!Slider->dummy  || !Slider->i_to_o || 
               !Slider->o_to_i || !Slider->prop) {
               errno = ENOMEM;
               if(Slider->dummy)
                  free(Slider->dummy);
               if(Slider->i_to_o)
                  free(Slider->i_to_o);
               if(Slider->o_to_i)
                  free(Slider->o_to_i);
               if(Slider->prop)
                  free(Slider->prop);
               return(-1);
           }
           memcpy(Slider->dummy, &(IcsRec->buffer_in[Slider->start]), x2 * sizeof(wchar_t));
           Slider->outsize = x2 * 3;

           (int)(*(Slider->lang))(IcsRec,Slider->dummy,x2,Slider->dummy,&(Slider->outsize),Slider->i_to_o,Slider->o_to_i,Slider->prop);

           Slider->dummy[Slider->outsize] = 0;
           Slider->processed = 1;

           if (Slider->next) {
               Slider = Slider->next;
               loop = 1;
           } else {
               break;
           }

         } while (loop == 1);
     }

    /* To fill the InpToOut */
    if (IcsRec->size_out == 0)
        IcsRec->size_out = IcsRec->size;

    for (i = 0; i < IcsRec->size; i++)
        IcsRec->SrcToTrgMap[i] = -1;
    for (i = 0; i < IcsRec->size; i++)
        IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0x03FF;
    for (i = 0, j = 0; i < IcsRec->size; i++) {
        if ((IcsRec->A_level[IcsRec->TrgToSrcMap[i]] & 0x80) == 0)
            continue;
        else {
            IcsRec->TrgToSrcMap[j] = IcsRec->TrgToSrcMap[i];
            j++;
        }
    }
    for (i = 0; i < IcsRec->size_out; i++)
        IcsRec->SrcToTrgMap[(IcsRec->TrgToSrcMap[i] & 0xff)] = i;
    for (i = 0; i < IcsRec->size; i++) {
        if ((IcsRec->A_level[i] & 0x80) == 0) {
            if ((ucb.outLev) == LTR){
                if(i>=1)
                IcsRec->SrcToTrgMap[i] = IcsRec->SrcToTrgMap[i - 1];
            } else {
                if(i<IcsRec->size-1)
                IcsRec->SrcToTrgMap[i] = IcsRec->SrcToTrgMap[i + 1];
            }
        }
    }
    for (i = 0; i < IcsRec->size_out; i++)
        IcsRec->TrgToSrcMap[i] = IcsRec->TrgToSrcMap[i] & 0x03FF;

    Slider = &Start;
    do {
        int             seem = 0;
        if ((Slider->processed == 1) && (Slider->lang) && (seem == 0)) {
            if (IcsRec->SrcToTrgMap[Slider->start] < IcsRec->SrcToTrgMap[Slider->end]) {
                /*
                 * This means we don't have to reverse the transformed buffer
                 * for langs other than BIDI 
                 */
                if (Slider->outsize > 0) {
                    size_t         *store_in = NULL, *store_out = NULL;
                    wchar_t        *store1, *store2;
                    int            index1,  index2,  diff, offset;

                    diff = Slider->outsize - Slider->end + Slider->start - 1;
                    IcsRec->buffer_out[IcsRec->size_out + diff] = 0;

                    if (Slider->end < IcsRec->size - 1) {
                        store_in = (size_t *) calloc((IcsRec->size - Slider->end), sizeof(size_t));
                        if(!store_in) {
                            errno = ENOMEM;
                            freeMultiLingual(Slider);
                            return(-1);
                        }
                        store_out = (size_t *) calloc((IcsRec->size_out - IcsRec->SrcToTrgMap[Slider->end]), sizeof(size_t));
                        if(!store_in) {
                            errno = ENOMEM;
                            free(store_in);
                            freeMultiLingual(Slider);
                            return(-1);
                        }
                        for (index1 = Slider->end + 1, index2 = 0; index1 < IcsRec->size; index1++, index2++)
                            store_in[index2] = IcsRec->SrcToTrgMap[index1];

                        for (index1 = IcsRec->SrcToTrgMap[Slider->end] + 1, index2 = 0; index1 < IcsRec->size_out; index1++, index2++)
                            store_out[index2] = IcsRec->TrgToSrcMap[index1];

                        store1 = (wchar_t *) calloc((IcsRec->size - Slider->end), sizeof(wchar_t));
                        if(!store1) {
                            errno = ENOMEM;
                            free(store_in);
                            free(store_out);
                            freeMultiLingual(Slider);
                            return(-1);
                        }
                        memcpy(store1,&IcsRec->buffer_out[IcsRec->SrcToTrgMap[Slider->end]+1],(IcsRec->size-Slider->end)*sizeof(wchar_t));

                        store2 = (wchar_t *) & IcsRec->buffer_out[IcsRec->SrcToTrgMap[Slider->start] + Slider->outsize];
                        memcpy(store2, store1, (IcsRec->size - Slider->end) * sizeof(wchar_t));
                        free(store_in);
                        free(store_out);
                        free(store1);
                    }
                    for (index1 = Slider->start, index2 = 0; index2 < (Slider->end - Slider->start + 1); index1++, index2++)
                        IcsRec->SrcToTrgMap[index1] = Slider->i_to_o[index2] + IcsRec->SrcToTrgMap[Slider->start];

                    if (Slider->end < IcsRec->size - 1) {
                        for (index2 = Slider->end + 1; index2 < IcsRec->size; index2++)
                            IcsRec->SrcToTrgMap[index2] = store_in[index2 - Slider->end - 1] + diff;
                    }
                    for (index1 = 0; index1 < Slider->outsize; index1++)
                        IcsRec->buffer_out[IcsRec->SrcToTrgMap[Slider->start] + index1] = Slider->dummy[index1];

                    for (index1 = Slider->start; index1 < Slider->end + 1; index1++) {
                        if ((Slider->prop[index1 - Slider->start] & 0x80) == 0)
                            IcsRec->A_level[index1] &= 0x0F;    /* Have the same level,
                                                                 * but is zero-width */
                    }

                    offset = IcsRec->TrgToSrcMap[IcsRec->SrcToTrgMap[Slider->start]];
                    for (index1 = IcsRec->SrcToTrgMap[Slider->start], index2 = 0; index2 < Slider->outsize; index1++, index2++) {
                        IcsRec->TrgToSrcMap[index1] = Slider->o_to_i[index2] + offset;
                    }

                    IcsRec->size_out += diff;
                    if (Slider->end < IcsRec->size - 1) {
                        for (index2 = index1; index2 < IcsRec->size_out; index2++)
                            IcsRec->TrgToSrcMap[index2] = store_out[index2 - index1];
                    }
                    seem = 1;
                }
            }
        }
        if (Slider->next) {
            Slider = Slider->next;
            seem = 0;
            loop = 1;
            continue;
        } else {
            loop = 0;
            continue;
        }
    } while (loop == 1);

    /* To free up the allocated memory */
    freeMultiLingual(Slider);

    return (RC);
}

/**************************************************************************/
/* FUNCTION NAME: specialChar()                                           */
/*                                                                        */
/* This function is used to find the special characters.                  */
/**************************************************************************/
int 
specialChar(wchar_t ch)
{
    if(( ch >= 0x065C && ch <= 0x065F) || (ch >=  0xFEED && ch <=  0xFEEE) || ch == 0x0648 || ch == 0x0624 )
      return(1);
    else if  (ch>0x064A && ch<0x0653)
      return(2);
    else
      return(0);
}


/*************************************************************************/
/* FUNCTION NAME: getLang()                                              */
/*                                                                       */
/* This function is used for the Thai and Vietnamese languages.          */
/*************************************************************************/
int 
getLang(pMultiLingual Pointer, int index, PICSPARAMRec IcsRec)
{
    MultiLingual    Start = { 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
    int             flag  = 0;
    int             rc    = 0;
    wchar_t         xChar;

    Pointer->start = index;
    Pointer->end   = index;
    for (Pointer->end = index; Pointer->end < IcsRec->size; Pointer->end++) {
        xChar = IcsRec->buffer_in[Pointer->end];
        if (xChar >= 0x0E00 && xChar <= 0x0E7F) {       /* Thai Language */
            if (flag == THAI)
                continue;
            else if (flag == 0) {
                Pointer->lang = ThaiTransform;
                Pointer->next = (pMultiLingual) calloc(1, sizeof(MultiLingual));
                if(!Pointer->next) {
                    errno = ENOMEM;
                    return(-1);
                }
                *(Pointer->next) = Start;
                flag = THAI;
                rc = 1;
            }
        }
         /* End of the Thai */ 
        else if (IsVN(xChar)) { /* Vietnamese  Language  */
            if (flag == VN || IsEnglish(xChar)){
                flag = VN;
                continue;
                }
            else if (flag == 0) {
                Pointer->lang = VNTransform;
                Pointer->next = (pMultiLingual) calloc(1, sizeof(MultiLingual));
                if(!Pointer->next) {
                    errno = ENOMEM;
                    return(-1);
                }
                *(Pointer->next) = Start;
                flag = VN;
                rc = 1;
            } else {
                Pointer->flag = flag;
                flag = 0;
                Pointer->end--;
                index = Pointer->end + 1;
                Pointer->next->prev = Pointer;
                Pointer = Pointer->next;
                Pointer->start = index;
                Pointer->end = index;
            }
        }
         /* End of the Vietnamese */ 
        else {  /* Bidi Languages */
            if (flag == BI || IsEnglish(xChar)){
                flag = BI;
                continue;
                }
            else if (flag == 0) {
                Pointer->lang = NULL;
                flag = BI;
                Pointer->next = (pMultiLingual) calloc(1, sizeof(MultiLingual));
                if(!Pointer->next) {
                    errno = ENOMEM;
                    return(-1);
                }
                *(Pointer->next) = Start;
            } else {
                Pointer->flag = flag;
                flag = 0;
                Pointer->end--;
                index = Pointer->end + 1;
                Pointer->next->prev = Pointer;
                Pointer = Pointer->next;
                Pointer->start = index;
                Pointer->end = index;
            }
        }                       /* End of Bidi */
    }
    /* To solve the extra end character problem */
    if (Pointer->end >= IcsRec->size)
        Pointer->end = IcsRec->size - 1;
    return (rc);
}


/*************************************************************************/
/* FUNCTION NAME: IsVN()                                                 */
/*                                                                       */
/* This function is used for the Vietnamese languages.                   */
/*************************************************************************/
BooleanValue 
IsVN(wchar_t xChar)
{
    if (xChar == 0x0102 || xChar == 0x0103 || xChar == 0x0110 || xChar == 0x0111 || xChar == 0x0128 || xChar == 0x0129  || 
	xChar == 0x0152 || xChar == 0x0153 || xChar == 0x0168 || xChar == 0x0169 || xChar == 0x0178 || xChar == 0x01a0  || 
	xChar == 0x01a1 || xChar == 0x01af || xChar == 0x01b0 || xChar == 0x0300 || xChar == 0x0301 || xChar == 0x0303  || 
	xChar == 0x0309 || xChar == 0x0323 || xChar == 0x0340 || xChar == 0x0341 || (xChar >= 0x1ea0 && xChar <= 0x1ef9)|| 
	xChar == 0x20ab)
        return (TRUE);
    return (FALSE);
}

/*************************************************************************/
/* FUNCTION NAME: IsEnglish()                                            */
/*                                                                       */
/* This function is used for the English languages.                      */
/*************************************************************************/
BooleanValue 
IsEnglish(short xChar)
{
    if ( ( xChar >= 0x0000 ) && ( xChar <= 0x00ff ) )
        return (TRUE);
    return (FALSE);
}

/*************************************************************************/
/* FUNCTION NAME: freeMultiLingual()                                     */
/*                                                                       */
/* This function is used to free allocated memory.                       */
/*************************************************************************/
int 
freeMultiLingual(pMultiLingual FreeMe)
{
    void            *tempp;
    int             next = 0;
    do {
        tempp = FreeMe;
        if (FreeMe->dummy)
            free(FreeMe->dummy);
        if (FreeMe->i_to_o)
            free(FreeMe->i_to_o);
        if (FreeMe->o_to_i)
            free(FreeMe->o_to_i);
        if (FreeMe->prop)
            free(FreeMe->prop);
        if (FreeMe->next)
            free(FreeMe->next);
        if (FreeMe->prev) {
            next = 1;
            FreeMe = FreeMe->prev;
            free(tempp);
        } else {
            next = 0;
        }
    } while (next == 1);
}

/***************************************************************************************/
/* The type of each character is found in these two tables MasterTable & ExtendedTable */
/***************************************************************************************/
const unsigned char MasterTable[GRP_SIZE] =
{
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
  /* 0- */ Table0, UBAT_L, UBAT_L, UBAT_L, UBAT_L, Table1, Table2, UBAT_R, UBAT_R, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 1- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* 2- */ Table3, Table4, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* 3- */ Table5, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 4- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 5- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 6- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 7- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 8- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* 9- */ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
  /* A- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* B- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* C- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* D- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* E- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
  /* F- */ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_L, UBAT_L, Table6, UBAT_R, UBAT_R, Table7, Table8
};

/****************************************************************************************************************************************/

#define NUMBER_OF_TABLES       9

const unsigned char      ExtendedTable[NUMBER_OF_TABLES][GRP_SIZE] =
{
 { 
/* Table 0: Unicode 00xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_S, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_W, UBAT_N, UBAT_N, UBAT_ET,UBAT_ET,UBAT_ET,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,UBAT_ET,UBAT_CS,UBAT_ET,UBAT_ES,UBAT_ES,
/*3-*/ UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_CS,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*5-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*7-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*8-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*9-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*A-*/ UBAT_W, UBAT_N, UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*B-*/ UBAT_ET,UBAT_ET,UBAT_EN,UBAT_EN,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_EN,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*C-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*D-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*E-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*F-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L
 },
 {
/* Table 1: Unicode 05xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*3-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*4-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*5-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*6-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*7-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*8-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*9-*/ UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*A-*/ UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*B-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*C-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*D-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*E-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*F-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N
 },
 {                     
/* Table 2: Unicode 06xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_R, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*3-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*5-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_AN,UBAT_ET,UBAT_AN,UBAT_AN,UBAT_R, UBAT_N, UBAT_N,
/*7-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*8-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*9-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*A-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*B-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N,
/*C-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N,
/*D-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*E-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N,
/*F-*/ UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N
 },
 {                              
/* Table 3: Unicode 20xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_W, UBAT_W, UBAT_W, UBAT_W, UBAT_W, UBAT_W, UBAT_W, UBAT_ES,UBAT_W, UBAT_W, UBAT_W, UBAT_W, UBAT_N, UBAT_N, UBAT_L, UBAT_R,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_B, UBAT_B, UBAT_BD, UBAT_N, UBAT_N,UBAT_N,UBAT_BD, UBAT_N,
/*3-*/ UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*5-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,UBAT_BD,UBAT_BD,UBAT_BD,UBAT_BD,
/*7-*/ UBAT_EN,UBAT_N, UBAT_N, UBAT_N,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_ET,UBAT_ET, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*8-*/ UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_ET,UBAT_ET,UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*9-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*A-*/ UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_ET,UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*B-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*C-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*D-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*E-*/ UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*F-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N
 },
 {                              
/* Table 4: Unicode 21xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*3-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*5-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*7-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*8-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*9-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*A-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*B-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*C-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*D-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*E-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*F-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N
 },
 {                              
/* Table 5: Unicode 30xx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_W, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*3-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*5-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*6-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*7-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*8-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*9-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N,
/*A-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*B-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*C-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*D-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*E-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*F-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N
 },
 {                              
/* Table 6: Unicode FBxx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_R, UBAT_R,
/*2-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*3-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_N,
/*4-*/ UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*5-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*6-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*7-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*8-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*9-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*A-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*B-*/ UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*C-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*D-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*E-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*F-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R
 },
 {
/* Table 7: Unicode FExx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*2-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*3-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*5-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*7-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_R, UBAT_N, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*8-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*9-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*A-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*B-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*C-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*D-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*E-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R,
/*F-*/ UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_R, UBAT_N, UBAT_N, UBAT_N
 },
 {                              
/* Table 8: Unicode FFxx ****************************************************************************************************************/
/*         0       1       2       3       4       5       6       7       8       9       A       B       C       D       E       F    */
/****************************************************************************************************************************************/
/*0-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*1-*/ UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN,UBAT_EN, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,UBAT_N,
/*2-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*3-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*4-*/ UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*5-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*6-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*7-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*8-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*9-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*A-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*B-*/ UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N,
/*C-*/ UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L,
/*D-*/ UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_L, UBAT_L, UBAT_L, UBAT_N, UBAT_N, UBAT_N,
/*E-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N,
/*F-*/ UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N, UBAT_N
  }
};
