/***********************************************************



((c) Copyright 2002 IBM. All Rights Reserved)." Licensed by IBM under the X license."

 COPYRIGHT AND PERMISSION NOTICE


Copyright (c) 2002 IBM Corporation

All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, provided that the above
copyright notice(s) and this permission notice appear in all copies of
the Software and that both the above copyright notice(s) and this
permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder
shall not be used in advertising or otherwise to promote the sale, use
or other dealings in this Software without prior written authorization
of the copyright holder.

X Window System is a trademark of The Open Group

******************************************************************/
                                                       
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <langinfo.h>
#include <iconv.h>
#include <errno.h>
#include <sys/errno.h>
#include "uc_bdstruct.h"
#include "layout_int.h"
#include "layout_backward.h"
#include "arabic.h"
#include "uc_data.h"

#ifndef MULTIPLIER
#define MULTIPLIER 2
#endif


extern int Invertbuffer(
      unsigned int *buffer, size_t lower_limit, size_t upper_limit);

/***************************************************************************/
/* FUNTCION NAME: wcsBidiTransform()                                       */
/*                                                                         */
/* This function is the one called from the m_wtransform_layout()          */
/* interface of the X/Open PLS APIs                                        */
/* It is responsible for the transformation of wide char code sets.        */
/***************************************************************************/                        
int 
wcsBidiTransform(LayoutObject context,
                 wchar_t * InpBuf,
                 size_t InpSize,
                 void *OutBuf,
                 size_t * OutSize,
                 size_t * ToOutBuf,
                 size_t * ToInpBuf,
                 unsigned char *BidiLvl,
                 size_t * InpBufIndex)
{
    size_t         *InpToOut     = NULL, *OutToInp = NULL;
    size_t         innum         = 0,    outnum    = 0;
    int            num_flag;
    int            txtShp_flag   = 0;
    int            outputIndex   = 0;
    int            index         = 0;
    int            needToInvert  = FALSE;
    int            i, x, buflen  = 0;
    int            RC = 0, rc    = 0;
    int            strongChar;
    ICSPARAMRec    IcsRec;
    LayoutCore     Values;
    char           *inpconvbuf   = NULL;
    char           *outconvbuf   = NULL;
    char           *outpbuf      = NULL;
    wchar_t        *tempbuf      = NULL;
    wchar_t        *temptrg      = NULL;
    Default        defaultData;
    Arabic         arabicData;
    UbData         uniData;

    Values      = &context->core;
    defaultData = context->private_data;
    arabicData  = defaultData->locale_data;
    uniData     = arabicData->uni_data;

    IcsRec.A_level = NULL;
    /********************* Start of String Conversion ****************/
    /* Determination of the length srcLen of source text to process  */ 
    /*****************************************************************/
    if((InpBuf == (wchar_t *)NULL) || ( *InpBuf == (wchar_t)NULL)) {
        errno = EINVAL;
        RC = -1;
        return(RC);
    }
    if (OutSize == NULL || InpSize == 0) {
        errno = EINVAL;
        RC = -1;
        return (RC);
    }
    /* decide on length of input string if not given */
    if (InpSize == -1) {
        i = 0;
        while (InpBuf[i] > (wchar_t) NULL) {
            if (i >= *OutSize && *OutSize) {
                RC = RC | 1;
                break;
            }
            i++;
        }
        InpSize = i;
    }
    if (*OutSize == 0) {/* calculate size of output buffer and return */
        *OutSize = InpSize;
    }
    if (InpSize > *OutSize) {
        InpSize = *OutSize;
        RC = RC | 1;
        errno = E2BIG;
    }

    /* decide on numerical shape flag */
    if (Values->numerals->inp != Values->numerals->out)
        num_flag = Values->numerals->out;
    else
        num_flag = NUMERALS_NONE;

    strongChar = StrongChar(InpBuf, InpSize);

    /* decide on intial shaping direction */
    if (Values->type_of_text->inp == TEXT_IMPLICIT) {
        /* for implicit source strings, shaping is always RTL */
        if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_in = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->inp == CONTEXT_LTR) {
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (Values->context->inp == CONTEXT_RTL) {
                IcsRec.orient_in = ORIENTATION_RTL;
            } else {
                IcsRec.orient_in = ORIENTATION_LTR;
            }
        } else
            IcsRec.orient_in = Values->orientation->inp;
        if ((Values->orientation->out == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_out = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_out = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->out == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->out == CONTEXT_LTR) {
                IcsRec.orient_out = ORIENTATION_LTR;
            } else if (Values->context->out == CONTEXT_RTL) {
                IcsRec.orient_out = ORIENTATION_RTL;
            } else {
                IcsRec.orient_out = ORIENTATION_LTR;
            }
        } else {
            IcsRec.orient_in  = Values->orientation->inp;
            IcsRec.orient_out = Values->orientation->out;
        }
    } else {  /* for visual source string, shaping is according to string orientation */
        if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
            (strongChar != 0)) {
            if (strongChar == 1) {  /* Means that the first strong char is LTR */
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (strongChar == 2) {  /* The first strong char is RTL */
                IcsRec.orient_in = ORIENTATION_RTL;
            }
        } else if ((Values->orientation->inp == ORIENTATION_CONTEXTUAL) && 
                   (strongChar == 0)) {
            if (Values->context->inp == CONTEXT_LTR) {
                IcsRec.orient_in = ORIENTATION_LTR;
            } else if (Values->context->inp == CONTEXT_RTL) {
                IcsRec.orient_in = ORIENTATION_RTL;
            } else {
                IcsRec.orient_in = ORIENTATION_LTR;
            }
        } else {
            IcsRec.orient_in = Values->orientation->inp;
            IcsRec.orient_out = Values->orientation->out;
        }
    }
    /*************************************************************************/
    buflen = InpSize;

    temptrg = calloc(InpSize * MULTIPLIER, sizeof(wchar_t));
    if (!temptrg) {
        errno = ENOMEM;
        return (-1);
    }
    tempbuf = calloc(InpSize * MULTIPLIER, sizeof(wchar_t));
    if (!tempbuf) {
        errno = ENOMEM;
        free(temptrg);
        return (-1);
    }

    InpToOut = calloc(InpSize * MULTIPLIER, sizeof(size_t));
    if (!InpToOut) {
        errno = ENOMEM;
        free(temptrg);
        free(tempbuf);
        return (-1);
    }

    OutToInp = calloc(InpSize * MULTIPLIER, sizeof(size_t));
    if (!OutToInp) {
        errno = ENOMEM;
        free(temptrg);
        free(tempbuf);
        free(InpToOut);
        return (-1);
    }

    memset(temptrg,0,InpSize * MULTIPLIER * sizeof(wchar_t));
    memset(InpToOut,0,InpSize * MULTIPLIER * sizeof(wchar_t));
    memset(OutToInp,0,InpSize * MULTIPLIER * sizeof(wchar_t));

    memcpy(temptrg, InpBuf, InpSize * sizeof(wchar_t));
    /* InpBufIndex offset to InpBuf */
    if (InpBufIndex) {
        if ((*InpBufIndex > Values->shape_context_size->back))
            memcpy(tempbuf, InpBuf + *InpBufIndex, InpSize * sizeof(wchar_t));
        else
            memcpy(tempbuf, InpBuf, InpSize * sizeof(wchar_t));
    } else
        memcpy(tempbuf, InpBuf, InpSize * sizeof(wchar_t));

    /*-------------------------------------------------*/
    /* The following fills the IcsRec struct variables */
    /*-------------------------------------------------*/
    index = 0;
    if (Values->type_of_text->inp != Values->type_of_text->out){
        needToInvert=(Values->type_of_text->inp == TEXT_VISUAL)&&
                     (IcsRec.orient_out != IcsRec.orient_in);
        IcsRec.order_flag = TRUE;
	if(arabicData->SpecialSh->out == TEXT_SPECIAL){
	    IcsRec.compc    = TRUE;
	    IcsRec.backward = TRUE;
	} else {
	    IcsRec.compc    = defaultData->compc;
	    IcsRec.backward = FALSE;
	}
        IcsRec.wordbreak = arabicData->Word->out;
        IcsRec.codeset = Values->shape_charset;
        IcsRec.num_flag = num_flag;
        if (Values->text_shaping->inp != Values->text_shaping->out)
            IcsRec.txtShp_flag = Values->text_shaping->out;
        else
            IcsRec.txtShp_flag = FALSE;

        IcsRec.formShp  = 0;
        IcsRec.boc_flag = 0;
        if (Values->type_of_text->inp != Values->type_of_text->out)
            IcsRec.flip_flag = TRUE;
        else
            IcsRec.flip_flag = FALSE;
        IcsRec.symmetric = 
              (Values->swapping->inp != Values->swapping->out);
	if(InpBufIndex)
	    IcsRec.size   = buflen - *InpBufIndex;
	else
	    IcsRec.size   = buflen;
        IcsRec.buffer_in  = (PUNI_CHAR) temptrg;
        IcsRec.buffer_out =  OutBuf;
        if (Values->type_of_text->out == TEXT_VISUAL)
             IcsRec.Tables = TEXT_VISUAL;
        else
             IcsRec.Tables = TEXT_IMPLICIT;
        if (BidiLvl)
             IcsRec.A_level = &BidiLvl[index];
        else {
             IcsRec.A_level = calloc(InpSize, sizeof(char));
             if(!IcsRec.A_level) {
                 errno = ENOMEM;
                 free(temptrg);
                 free(tempbuf);
                 free(InpToOut);
                 free(OutToInp);
                 return(-1);
             }
        }
        IcsRec.SrcToTrgMap = (size_t*) &InpToOut[index];
        IcsRec.TrgToSrcMap = (size_t*) &OutToInp[outputIndex];
       
        if(needToInvert) {
            IcsRec.orient_in = IcsRec.orient_out;
            if((inverse((char *)IcsRec.buffer_in,IcsRec.size,sizeof(wchar_t))) == -1) {
                if(errno == ENOMEM) {
                    free(temptrg);
                    free(tempbuf);
                    free(InpToOut);
                    free(OutToInp);
                    if (!BidiLvl && IcsRec.A_level) 
                         free(IcsRec.A_level);
                    return(-1);
                }
            }
        }

        /* Now call the actual transformation process from the function UCB2VIS */
	/* which will call from within the 5 passes from pass1 to pass5         */               
        rc = UCB2VIS(&IcsRec);

        if(rc == -1 && errno == ENOMEM) {
            free(temptrg);
   	    free(tempbuf);
            free(InpToOut);
            free(OutToInp);
            if (!BidiLvl && IcsRec.A_level) 
                 free(IcsRec.A_level);
            return(-1);
        }

        if(needToInvert) {
           if(IcsRec.SrcToTrgMap)
              Invertbuffer(IcsRec.SrcToTrgMap,0,IcsRec.size-1);
           if(IcsRec.TrgToSrcMap)
              Invertbuffer(IcsRec.TrgToSrcMap,0,IcsRec.size-1);
        }
        if (IcsRec.size == 1)
            IcsRec.A_level[0] += 0x80;
        RC = rc | RC;
    } else {
            IcsRec.size        = buflen;
            IcsRec.size_out    = IcsRec.size;
            IcsRec.buffer_in   = (PUNI_CHAR) & tempbuf[index];
            IcsRec.buffer_out  = (PUNI_CHAR) & temptrg[outputIndex];
            IcsRec.SrcToTrgMap = &InpToOut[index];
            IcsRec.TrgToSrcMap = &OutToInp[outputIndex];

            for (i = 0; i < IcsRec.size; i++)
                IcsRec.TrgToSrcMap[i] = i;
            for (i = 0; i < IcsRec.size; i++)
                IcsRec.SrcToTrgMap[i] = i;
            if (BidiLvl)
                for (i = 0; i < IcsRec.size; i++)
                    BidiLvl[i] = 0x80;
            IcsRec.orient_out = Values->orientation->out;
            IcsRec.orient_in = Values->orientation->inp;
            needToInvert = (Values->type_of_text->inp == TEXT_VISUAL)&&
                           (IcsRec.orient_out != IcsRec.orient_in);
            if (needToInvert) {
                /* Inversion is required */
                if((inverse(IcsRec.buffer_out,
                    IcsRec.size, sizeof(wchar_t))) == -1) {
                    if(errno == ENOMEM) {
                        free(temptrg);
                        free(tempbuf);
                        free(InpToOut);
                        free(OutToInp);
                        return(-1);
                    }
                }
                if (ToInpBuf)
                    if((inverse(IcsRec.TrgToSrcMap,
                        IcsRec.size, sizeof(size_t))) == -1) {
                        if(errno == ENOMEM) {
                            free(temptrg);
                            free(tempbuf);
                            free(InpToOut);
                            free(OutToInp);
                            return(-1);
                        }
                    }
                if (ToOutBuf)
                    for (i = 0; i < InpSize; i++)
                        IcsRec.SrcToTrgMap[i] = 
                                InpSize - 1 - IcsRec.SrcToTrgMap[i];
            } else {
                if(ToInpBuf)
                    for (i = 0; i < InpSize; i++)
                        ToInpBuf[i] = i;
                if(ToOutBuf)
                    for (i = 0; i < InpSize; i++)
                        ToOutBuf[i] = i;
            }
            memcpy(OutBuf,temptrg,InpSize*sizeof(wchar_t));
        }

        /********************************************************************/
        for (i = 0; i < IcsRec.size; i++)
            IcsRec.TrgToSrcMap[i] = IcsRec.TrgToSrcMap[i] & 0x00FF;
    /***********************Source To Target *******************************/
    if (ToOutBuf) {
        for (i = 0; i < InpSize; i++)
            ToOutBuf[i] = 0;
        i = 0;
        x = 0;
        while (i < InpSize) {
            ToOutBuf[i] = IcsRec.SrcToTrgMap[x];
            i++;
            x++;
        }
    }
    /***********************************************************************/
    /* copy temporary buffer to output buffer */
    if (OutBuf) {
        if ((Values->text_shaping->inp != Values->text_shaping->out)
            && context->core.active_shape_editing
            && *OutSize != 0
            && !((Values->type_of_text->inp == TEXT_IMPLICIT) || 
                 (Values->type_of_text->out == TEXT_IMPLICIT))) {
            iconv_t         to_1046;
            iconv_t         to_UCS;
            char            *outputBuffer; 
            unsigned char   *inputBuffer = (unsigned char*) IcsRec.buffer_out;
            char            *outputPtr   = outputBuffer;
            size_t          inputSize    = IcsRec.size_out * sizeof(wchar_t);
            size_t          outputSize   = IcsRec.size_out;
        
            outputBuffer = (char *) calloc(IcsRec.size_out, sizeof(char)); 
            if(!outputBuffer) {
                errno = ENOMEM;
                free(temptrg);
                free(tempbuf);
                free(InpToOut);
                free(OutToInp);
                if (!BidiLvl && IcsRec.A_level)
                    free(IcsRec.A_level);
                return(-1);
            }
            to_1046 = iconv_open("IBM1046", "UCS-4");
            to_UCS  = iconv_open("UCS-4", "IBM1046");

	    /* A call to iconv is needed here to change from UCS-4 to 1046 */
            iconv(to_1046, (char **) &inputBuffer, &inputSize, &outputPtr, &outputSize); 

            outputSize = IcsRec.size_out - outputSize;

            iconv_close(to_1046);

            BidiShape(outputBuffer, InpSize, arabicData->OneCell->out,
                      arabicData->SpecialSh->out, 1,
                      Values->text_shaping->out);

            inputBuffer     = outputBuffer;
            inputSize       = IcsRec.size_out;
            outputPtr       = OutBuf;
            IcsRec.size_out = outputSize;
            outputSize      = outputSize * sizeof(wchar_t);

	    /* A call to iconv is needed here to change from 1046 to UCS-4 */
            iconv(to_UCS, (char **)&inputBuffer, &inputSize, &outputPtr, &outputSize); 

	    /* Close previously used iconv handles */
            iconv_close(to_UCS);
            iconv_close(to_1046);
            free(outputBuffer);
        }
        /*********Target To Source *********************/
        if (ToInpBuf) {
            for (i = 0; i < *OutSize; i++)
                ToInpBuf[i] = 0;
            i = 0;
            x = 0;
            while (i < *OutSize) {
                ToInpBuf[i] = OutToInp[x];
                i++;
                x++;
            }
        }
        /********************************************************/
    }

    free(temptrg);
    free(tempbuf);
    if(!BidiLvl && IcsRec.A_level)
       free(IcsRec.A_level);
    free(InpToOut);
    free(OutToInp);
    /********************* End of String Conversion *************************/

    return (RC);                /* successfully performed */
}


/*******************************************************************/
/* FUNCTION NAME: reverseEndian()                                  */
/*                                                                 */
/* This function handles the endian status and converts from little*/
/* endian to big endian in wide characters.                        */
/*******************************************************************/                                                                       
